from typing import Any
from typing import Dict
from typing import List


async def search_raw(
    hub,
    ctx,
    filters: List = None,
    resource_id: str = None,
) -> Dict:
    """
    Fetch one or more VPC peering connections from AWS.
    The return will be in the same format as what the boto3 api returns.

    Args:
        resource_id(string, optional): AWS VPC peering connection id to identify the resource.
        filters(list, optional): One or more filters: for example, tag :<key>, tag-key.
            A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_vpc_peering_connections

    Returns:
        {"result": True|False, "comment": A message List, "ret": Dict}

    """
    result = dict(comment=[], ret=None, result=True)
    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    if not syntax_validation["result"]:
        result["comment"] = list(syntax_validation["comment"])
        return result
    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )
    ret = await hub.exec.boto3.client.ec2.describe_vpc_peering_connections(
        ctx,
        Filters=boto3_filter,
        VpcPeeringConnectionIds=[resource_id] if resource_id else None,
    )
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]

    return result


async def update_vpc_peering_connection_options(
    hub,
    ctx,
    vpc_peering_connection_id: str,
    name,
    initial_options_ret: Dict[str, Any],
    new_options: Dict[str, bool],
) -> Dict[str, Any]:
    result = dict(comment=(), result=True, ret=None)

    # Create the existing options' dictionary by containing
    # only the options properties and not the resource_id, name etc.
    existing_options = {
        key: initial_options_ret["ret"][key]
        for key in initial_options_ret["ret"].keys() & new_options.keys()
    }

    if not new_options or all(value is None for value in new_options.values()):
        return result

    else:
        if None in new_options.values():
            for key, value in new_options.items():
                if value is None:
                    new_options[key] = existing_options[key]

        if new_options != existing_options:
            if not ctx.get("test", False):
                modify_options_ret = await hub.exec.boto3.client.ec2.modify_vpc_peering_connection_options(
                    ctx,
                    **{
                        "AccepterPeeringConnectionOptions": {
                            "AllowDnsResolutionFromRemoteVpc": new_options[
                                "peer_allow_remote_vpc_dns_resolution"
                            ],
                            "AllowEgressFromLocalClassicLinkToRemoteVpc": new_options[
                                "peer_allow_classic_link_to_remote_vpc"
                            ],
                            "AllowEgressFromLocalVpcToRemoteClassicLink": new_options[
                                "peer_allow_vpc_to_remote_classic_link"
                            ],
                        },
                        "RequesterPeeringConnectionOptions": {
                            "AllowDnsResolutionFromRemoteVpc": new_options[
                                "allow_remote_vpc_dns_resolution"
                            ],
                            "AllowEgressFromLocalClassicLinkToRemoteVpc": new_options[
                                "allow_classic_link_to_remote_vpc"
                            ],
                            "AllowEgressFromLocalVpcToRemoteClassicLink": new_options[
                                "allow_vpc_to_remote_classic_link"
                            ],
                        },
                        "VpcPeeringConnectionId": vpc_peering_connection_id,
                    },
                )

                if not modify_options_ret["result"]:
                    result["comment"] = modify_options_ret["comment"]
                    result["result"] = False
                    return result

            if result["result"]:
                # See which values are different in the dictionary and add those in ret property
                result["ret"] = {
                    key: new_options[key]
                    for key in existing_options
                    if new_options[key] != existing_options[key]
                }

        if ctx.get("test", False):
            result[
                "comment"
            ] = hub.tool.aws.comment_utils.would_update_resource_options_comment(
                "aws.ec2.vpc_peering_connection_options",
                vpc_peering_connection_id,
                name,
            )
        else:
            result[
                "comment"
            ] = hub.tool.aws.comment_utils.update_resource_options_comment(
                "aws.ec2.vpc_peering_connection_options",
                vpc_peering_connection_id,
                name,
            )

    return result


async def get_vpc_peering_connection_raw(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    resource_type: str = None,
    filters: List = None,
) -> Dict:
    """
    Get the raw vpc peering connection resource.

    Args:
        name(string): The name of the Idem state.
        resource_id(string, optional): AWS VPC peering connection id to identify the resource.
        resource_type: The type of the resource.
        filters(list, optional): One or more filters: for example, tag :<key>, tag-key.
            A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_vpc_peering_connections

    """
    result = dict(comment=[], ret=None, result=True)
    vpc_peering_connection_ret = (
        await hub.tool.aws.ec2.vpc_peering_connection_utils.search_raw(
            ctx=ctx,
            resource_id=resource_id,
            filters=filters,
        )
    )
    if not vpc_peering_connection_ret["result"]:
        if "InvalidVpcPeeringConnectionID.NotFound" in str(
            vpc_peering_connection_ret["comment"]
        ):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type=resource_type, name=name
                )
            )
            result["comment"] += list(vpc_peering_connection_ret["comment"])
            return result
        result["comment"] += list(vpc_peering_connection_ret["comment"])
        result["result"] = False
        return result
    if not vpc_peering_connection_ret["ret"]["VpcPeeringConnections"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type=resource_type, name=name
            )
        )
        return result

    resource = vpc_peering_connection_ret["ret"]["VpcPeeringConnections"][0]
    if len(vpc_peering_connection_ret["ret"]["VpcPeeringConnections"]) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_type=resource_type, resource_id=resource_id
            )
        )

    result["ret"] = resource

    return result



"""State module for managing AWS Launch Configurations."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    image_id: str = None,
    key_name: str = None,
    security_groups: List[str] = None,
    classic_link_vpc_id: str = None,
    classic_link_vpc_security_groups: List[str] = None,
    user_data: str = None,
    instance_id: str = None,
    instance_type: str = None,
    kernel_id: str = None,
    ramdisk_id: str = None,
    block_device_mappings: List[
        make_dataclass(
            "BlockDeviceMapping",
            [
                ("DeviceName", str),
                ("VirtualName", str, field(default=None)),
                (
                    "Ebs",
                    make_dataclass(
                        "Ebs",
                        [
                            ("SnapshotId", str, field(default=None)),
                            ("VolumeSize", int, field(default=None)),
                            ("VolumeType", str, field(default=None)),
                            ("DeleteOnTermination", bool, field(default=None)),
                            ("Iops", int, field(default=None)),
                            ("Encrypted", bool, field(default=None)),
                            ("Throughput", int, field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
                ("NoDevice", bool, field(default=None)),
            ],
        )
    ] = None,
    instance_monitoring: make_dataclass(
        "InstanceMonitoring", [("Enabled", bool, field(default=None))]
    ) = None,
    spot_price: str = None,
    iam_instance_profile: str = None,
    ebs_optimized: bool = None,
    associate_public_ip_address: bool = None,
    placement_tenancy: str = None,
    metadata_options: make_dataclass(
        "InstanceMetadataOptions",
        [
            ("HttpTokens", str, field(default=None)),
            ("HttpPutResponseHopLimit", int, field(default=None)),
            ("HttpEndpoint", str, field(default=None)),
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""Creates an AWS Launch Configuration.

    If you exceed your maximum limit of launch configurations, the call fails. To query this limit, call the
    DescribeAccountLimits API. For information about updating this limit, see Amazon EC2 Auto Scaling service quotas in
    the Amazon EC2 Auto Scaling User Guide.

    For more information, see Launch configurations in the Amazon EC2 Auto Scaling User Guide.

    Args:
        name(str):
            The name of the launch configuration. This name must be unique per Region per account.

        resource_id(str, Optional):
            AWS Launch Configuration name.

        image_id(str, Optional):
            The ID of the Amazon Machine Image (AMI) that was assigned during registration. For more
            information, see Finding an AMI in the Amazon EC2 User Guide for Linux Instances. Defaults to None.

            If you do not specify instance_id, you must specify image_id.

        key_name(str, Optional):
            The name of the key pair. For more information, see Amazon EC2 Key Pairs in the Amazon EC2 User
            Guide for Linux Instances. Defaults to None.

        security_groups(list[str], Optional):
            A list that contains the security groups to assign to the instances in the Auto Scaling group.
            For more information, see Control traffic to resources using security groups in the Amazon Virtual Private
            Cloud User Guide. Defaults to None.

        classic_link_vpc_id(str, Optional):
            The ID of a ClassicLink-enabled VPC to link your EC2-Classic instances to. For more information,
            see ClassicLink in the Amazon EC2 User Guide for Linux Instances and Linking EC2-Classic
            instances to a VPC in the Amazon EC2 Auto Scaling User Guide. This parameter can only be used if
            you are launching EC2-Classic instances. Defaults to None.

        classic_link_vpc_security_groups(list[str], Optional):
            The IDs of one or more security groups for the specified ClassicLink-enabled VPC.
            If you specify the classic_link_vpc_id parameter, you must specify this parameter. Defaults to None.

        user_data(str, Optional):
            The user data to make available to the launched EC2 instances. For more information, see Instance metadata
            and user data (Linux) and Instance metadata and user data (Windows). If you are using a command line tool,
            base64-encoding is performed for you, and you can load the text from a file. Otherwise, you must provide
            base64-encoded text. User data is limited to 16 KB. Defaults to None.

        instance_id(str, Optional):
            The ID of the instance to use to create the launch configuration. The new launch configuration derives
            attributes from the instance, except for the block device mapping.

            To create a launch configuration with a block device mapping or override any other instance attributes,
            specify them as part of the same request.

            For more information, see Creating a launch configuration using an EC2 instance in the Amazon EC2 Auto
            Scaling User Guide.

            If you do not specify instance_id, you must specify both image_id and instance_type. Defaults to None.

        instance_type(str, Optional):
            Specifies the instance type of the EC2 instance. For information about available instance types,
            see Available Instance Types in the Amazon EC2 User Guide for Linux Instances.

            If you do not specify instance_id, you must specify instance_type. Defaults to None.

        kernel_id(str, Optional):
            The ID of the kernel associated with the AMI. Defaults to None.

        ramdisk_id(str, Optional):
            The ID of the RAM disk to select. Defaults to None.

        block_device_mappings(list[dict[str, Any]], Optional):
            A block device mapping, which specifies the block devices for the instance. You can specify
            virtual devices and EBS volumes. For more information, see Block Device Mapping in the Amazon
            EC2 User Guide for Linux Instances. Defaults to None.

            * VirtualName(str, Optional):
                The name of the virtual device (for example, ephemeral0). You can specify either VirtualName or Ebs,
                but not both.

            * DeviceName(str):
                The device name exposed to the EC2 instance (for example, /dev/sdh or xvdh). For more information, see
                Device Naming on Linux Instances in the Amazon EC2 User Guide for Linux Instances.

            * Ebs(dict[str, Any], Optional):
                Parameters used to automatically set up EBS volumes when an instance is launched. You can specify
                either VirtualName or Ebs, but not both.

                * SnapshotId(str, Optional):
                    The snapshot ID of the volume to use.
                    You must specify either a VolumeSize or a SnapshotId.

                * VolumeSize(int, Optional):
                    The volume size, in GiBs.

                    The following are the supported volumes sizes for each volume type:
                        * gp2 and gp3: 1-16,384
                        * io1: 4-16,384
                        * st1 and sc1: 125-16,384
                        * standard: 1-1,024

                    You must specify either a SnapshotId or a VolumeSize.
                    If you specify both SnapshotId and VolumeSize , the volume size must be equal or greater than the
                    size of the snapshot.

                * VolumeType(str, Optional):
                    The volume type. For more information, see Amazon EBS volume types in the Amazon EC2 User Guide for
                    Linux Instances.

                    Valid values: standard | io1 | gp2 | st1 | sc1 | gp3

                * DeleteOnTermination(bool, Optional):
                    Indicates whether the volume is deleted on instance termination. For Amazon EC2 Auto Scaling, the
                    default value is true.

                * Iops (int, Optional):
                    The number of input/output (I/O) operations per second (IOPS) to provision for the volume. For gp3
                    and io1 volumes, this represents the number of IOPS that are provisioned for the volume. For gp2
                    volumes, this represents the baseline performance of the volume and the rate at which the volume
                    accumulates I/O credits for bursting.

                    The following are the supported values for each volume type:
                        * gp3: 3,000-16,000 IOPS
                        * io1: 100-64,000 IOPS

                    For io1 volumes, we guarantee 64,000 IOPS only for Instances built on the Nitro System . Other
                    instance families guarantee performance up to 32,000 IOPS.

                    Iops is supported when the volume type is gp3 or io1 and required only when the volume type is io1.
                    (Not used with standard , gp2 , st1 , or sc1 volumes.)

                * Encrypted(bool, Optional):
                    Specifies whether the volume should be encrypted. Encrypted EBS volumes can only be attached to
                    instances that support Amazon EBS encryption. For more information, see Supported instance types.
                    If your AMI uses encrypted volumes, you can also only launch it on supported instance types.

                    .. Note::
                        If you are creating a volume from a snapshot, you cannot create an unencrypted volume from an
                        encrypted snapshot. Also, you cannot specify a KMS key ID when using a launch configuration.

                        If you enable encryption by default, the EBS volumes that you create are always encrypted,
                        either using the Amazon Web Services managed KMS key or a customer-managed KMS key, regardless
                        of whether the snapshot was encrypted.

                        For more information, see Use Amazon Web Services KMS keys to encrypt Amazon EBS volumes in the
                        Amazon EC2 Auto Scaling User Guide.

                * Throughput(int, Optional):
                    The throughput (MiBps) to provision for a gp3 volume.

            * NoDevice(bool, Optional):
                Setting this value to true suppresses the specified device included in the block device mapping of the
                AMI from being mapped to the specified device name at launch.

                If NoDevice is true for the root device, instances might fail the EC2 health check. In that case,
                Amazon EC2 Auto Scaling launches replacement instances.

                If you specify NoDevice, you cannot specify Ebs.

        instance_monitoring(dict[str, Any], Optional):
            Controls whether instances in this group are launched with detailed (true) or basic (false)
            monitoring.
            The default value is true (enabled).

            .. Warning::
                When detailed monitoring is enabled, Amazon CloudWatch generates metrics every minute and your account
                is charged a fee. When you disable detailed monitoring, CloudWatch generates metrics every 5 minutes.
                For more information, see Configure Monitoring for Auto Scaling Instances in the Amazon EC2 Auto
                Scaling User Guide.

            * Enabled(bool, Optional):
                If true, detailed monitoring is enabled. Otherwise, basic monitoring is enabled.

        spot_price(str, Optional):
            The maximum hourly price to be paid for any Spot Instance launched to fulfill the request. Spot
            Instances are launched when the price you specify exceeds the current Spot price. For more
            information, see Requesting Spot Instances in the Amazon EC2 Auto Scaling User Guide.

            Valid Range: Minimum value of 0.001

            .. Note::
                When you change your maximum price by creating a new launch configuration, running instances will
                continue to run as long as the maximum price for those running instances is higher than the current
                Spot price.

            Defaults to None.

        iam_instance_profile(str, Optional):
            The name or the Amazon Resource Name (ARN) of the instance profile associated with the IAM role for the
            instance. The instance profile contains the IAM role. For more information, see IAM role for applications
            that run on Amazon EC2 instances in the Amazon EC2 Auto Scaling User Guide. Defaults to None.

        ebs_optimized(bool, Optional):
            Specifies whether the launch configuration is optimized for EBS I/O (true) or not (false). The optimization
            provides dedicated throughput to Amazon EBS and an optimized configuration stack to provide optimal I/O
            performance. This optimization is not available with all instance types. Additional fees are incurred when
            you enable EBS optimization for an instance type that is not EBS-optimized by default. For more
            information, see Amazon EBS-optimized instances in the Amazon EC2 User Guide for Linux Instances.

            The default value is false.

        associate_public_ip_address(bool, Optional):
            Specifies whether to assign a public IPv4 address to the group's instances. If the instance is launched
            into a default subnet, the default is to assign a public IPv4 address, unless you disabled the option to
            assign a public IPv4 address on the subnet. If the instance is launched into a nondefault subnet, the
            default is not to assign a public IPv4 address, unless you enabled the option to assign a public IPv4
            address on the subnet.

            If you specify true , each instance in the Auto Scaling group receives a unique public IPv4 address. For
            more information, see Launching Auto Scaling instances in a VPC in the Amazon EC2 Auto Scaling User Guide.

            If you specify this property, you must specify at least one subnet for VPCZoneIdentifier when you create
            your group.

        placement_tenancy(str, Optional):
            The tenancy of the instance, either default or dedicated. An instance with dedicated tenancy runs on
            isolated, single-tenant hardware and can only be launched into a VPC. To launch dedicated instances into a
            shared tenancy VPC (a VPC with the instance placement tenancy attribute set to default), you must set
            the value of this parameter to dedicated. For more information, see Configuring instance tenancy with
            Amazon EC2 Auto Scaling in the Amazon EC2 Auto Scaling User Guide.

            If you specify PlacementTenancy, you must specify at least one subnet for VPCZoneIdentifier when you create
            your group.

            Valid Values: default | dedicated. Defaults to None.

        metadata_options(dict[str, Any], Optional):
            The metadata options for the instances. For more information, see Configuring the Instance
            Metadata Options in the Amazon EC2 Auto Scaling User Guide. Defaults to None.

            * HttpTokens(str, Optional):
                The state of token usage for your instance metadata requests. If the parameter is not specified in the
                request, the default state is `optional`.

                If the state is `optional`, you can choose to retrieve instance metadata with or without a signed token
                header on your request. If you retrieve the IAM role credentials without a token, the version 1.0 role
                credentials are returned. If you retrieve the IAM role credentials using a valid signed token, the
                version 2.0 role credentials are returned.

                If the state is `required`, you must send a signed token header with any instance metadata retrieval
                requests. In this state, retrieving the IAM role credentials always returns the version 2.0
                credentials; the version 1.0 credentials are not available.

            * HttpPutResponseHopLimit(int, Optional):
                The desired HTTP PUT response hop limit for instance metadata requests. The larger the number, the
                further instance metadata requests can travel.

                Default: 1

            * HttpEndpoint(str, Optional):
                This parameter enables or disables the HTTP metadata endpoint on your instances. If the parameter is
                not specified, the default state is enabled.

                .. Note::
                    If you specify a value of disabled, you will not be able to access your instance metadata.

    Request Syntax:
        .. code-block:: sls

            [launch_configuration-resource-id]:
              aws.autoscaling.launch_configuration.present:
                - name: 'string'
                - resource_id: 'string'
                - image_id: 'string'
                - key_name: 'string'
                - security_groups: List
                - classic_link_vpc_id: 'string'
                - classic_link_vpc_security_groups: List
                - user_data: 'string'
                - instance_type: 'string'
                - kernel_id: 'string'
                - ramdisk_id: 'string'
                - block_device_mappings: List
                - instance_monitoring: Dict
                - spot_price: 'string'
                - iam_instance_profile: 'string'
                - ebs_optimized: true/false
                - metadata_options: Dict
                - associate_public_ip_address: true/false

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem-test-launch_configuration:
              aws.autoscaling.launch_configuration.present:
                - name: idem-test-launch_configuration
                - resource_id: idem-test-launch_configuration
                - image_id: ami-73949613
                - key_name: test-n
                - security_groups:
                    - sg-00bbd7b8d31f6bcad
                    - sg-03064434ef6546344
                - classic_link_vpc_id: vpc-123456
                - classic_link_vpc_security_groups:
                    - sg-00bbd7b8d31f6abcd
                    - sg-03064434ef6541111
                - user_data: /etc/eks/bootstrap.sh
                - instance_type: t2.micro
                - kernel_id: aki-02b79b47
                - ramdisk_id: ari-177e2d52
                - block_device_mappings:
                    - DeviceName: /dev/sda1
                      Ebs:
                        DeleteOnTermination: false
                        Encrypted: false
                        SnapshotId: snap-0526fa8583b2ef05f
                        VolumeSize: 8
                        VolumeType: gp2
                    - DeviceName: /dev/sda
                      Ebs:
                        DeleteOnTermination: false
                        Encrypted: false
                        SnapshotId: snap-05cfe4be1b16a2e0f
                        VolumeSize: 8
                        VolumeType: gp2
                - instance_monitoring:
                    Enabled: true
                - spot_price: "0.045"
                - iam_instance_profile: arn:aws:iam::537227425989:instance-profile/aws-elasticbeanstalk-ec2-role
                - ebs_optimized: true
                - associate_public_ip_address: true
                - metadata_options:
                    HttpEndpoint: enabled
                    HttpPutResponseHopLimit: 64
                    HttpTokens: optional

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    launch_configuration_created = False
    before = None
    desired_state = {
        "name": name,
        "resource_id": name,
        "image_id": image_id,
        "key_name": key_name,
        "security_groups": security_groups,
        "classic_link_vpc_id": classic_link_vpc_id,
        "classic_link_vpc_security_groups": classic_link_vpc_security_groups,
        "user_data": user_data,
        "instance_id": instance_id,
        "instance_type": instance_type,
        "kernel_id": kernel_id,
        "ramdisk_id": ramdisk_id,
        "block_device_mappings": block_device_mappings,
        "instance_monitoring": instance_monitoring,
        "spot_price": spot_price,
        "iam_instance_profile": iam_instance_profile,
        "ebs_optimized": ebs_optimized,
        "associate_public_ip_address": associate_public_ip_address,
        "placement_tenancy": placement_tenancy,
        "metadata_options": metadata_options,
    }

    launch_conf_args = {
        "LaunchConfigurationName": name,
        "ImageId": image_id,
        "KeyName": key_name,
        "SecurityGroups": security_groups,
        "ClassicLinkVPCId": classic_link_vpc_id,
        "ClassicLinkVPCSecurityGroups": classic_link_vpc_security_groups,
        "UserData": user_data,
        "InstanceId": instance_id,
        "InstanceType": instance_type,
        "KernelId": kernel_id,
        "RamdiskId": ramdisk_id,
        "BlockDeviceMappings": block_device_mappings,
        "InstanceMonitoring": instance_monitoring,
        "SpotPrice": spot_price,
        "IamInstanceProfile": iam_instance_profile,
        "EbsOptimized": ebs_optimized,
        "AssociatePublicIpAddress": associate_public_ip_address,
        "PlacementTenancy": placement_tenancy,
        "MetadataOptions": metadata_options,
    }

    if resource_id:
        before = await hub.exec.boto3.client.autoscaling.describe_launch_configurations(
            ctx, LaunchConfigurationNames=[resource_id]
        )

    # Launch Configurations cannot be updated after creation with the Amazon Web Service API.
    if before and before["ret"].get("LaunchConfigurations"):  # resource present
        result[
            "old_state"
        ] = hub.tool.aws.autoscaling.conversion_utils.convert_raw_launch_configuration_to_present(
            raw_resource=before["ret"]["LaunchConfigurations"][0],
            idem_resource_name=name,
        )
        result["comment"] = (
            f"aws.autoscaling.launch_configuration '{name}' already exists",
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={}, desired_state=desired_state
            )
            result["comment"] = (
                f"Would create aws.autoscaling.launch_configuration '{name}'",
            )
            return result

        try:
            ret = await hub.exec.boto3.client.autoscaling.create_launch_configuration(
                ctx,
                **launch_conf_args,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result

            result["comment"] = (
                f"Created aws.autoscaling.launch_configuration '{name}'",
            )
            launch_configuration_created = True
            resource_id = name
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    try:
        if launch_configuration_created:
            after = (
                await hub.exec.boto3.client.autoscaling.describe_launch_configurations(
                    ctx, LaunchConfigurationNames=[resource_id]
                )
            )
            result[
                "new_state"
            ] = hub.tool.aws.autoscaling.conversion_utils.convert_raw_launch_configuration_to_present(
                raw_resource=after["ret"]["LaunchConfigurations"][0],
                idem_resource_name=name,
            )
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""Deletes the specified AWS Launch Configuration.

    The launch configuration must not be attached to an Auto Scaling group.
    When this call completes, the launch configuration is no longer available for use.

    Args:
        name(str): The name of the launch configuration.
        resource_id(str, Optional): AWS launch configuration name. Idem automatically considers this resource being absent if
         this field is not specified.

    Request Syntax:
        .. code-block:: sls

            [launch_configuration-resource-id]:
              aws.autoscaling.launch_configuration.absent:
                - name: 'string'
                - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem-test-launch_configuration:
              aws.autoscaling.launch_configuration.absent:
                - name: idem-test-launch_configuration
                - resource_id: idem-test-launch_configuration
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.autoscaling.launch_configuration", name=name
        )
        return result

    before = await hub.exec.boto3.client.autoscaling.describe_launch_configurations(
        ctx, LaunchConfigurationNames=[resource_id]
    )
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result
    if not before["ret"].get("LaunchConfigurations"):
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.autoscaling.launch_configuration", name=name
        )
    else:
        result[
            "old_state"
        ] = hub.tool.aws.autoscaling.conversion_utils.convert_raw_launch_configuration_to_present(
            raw_resource=before["ret"]["LaunchConfigurations"][0],
            idem_resource_name=name,
        )

        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.autoscaling.launch_configuration", name=name
            )
            return result

        try:
            ret = await hub.exec.boto3.client.autoscaling.delete_launch_configuration(
                ctx, LaunchConfigurationName=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.autoscaling.launch_configuration", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Lists AWS Launch Configurations in the account and Region.

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.


    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.autoscaling.launch_configuration
    """
    result = {}
    ret = await hub.exec.boto3.client.autoscaling.describe_launch_configurations(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe launch_configuration {ret['comment']}")
        return {}

    for launch_configuration in ret["ret"].get("LaunchConfigurations"):
        resource_id = launch_configuration.get("LaunchConfigurationName")
        resource_translated = hub.tool.aws.autoscaling.conversion_utils.convert_raw_launch_configuration_to_present(
            raw_resource=launch_configuration, idem_resource_name=resource_id
        )
        result[resource_id] = {
            "aws.autoscaling.launch_configuration.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



import boto3.exceptions
import botocore.errorfactory
import botocore.exceptions


def __init__(hub):
    for module in (botocore.errorfactory, botocore.exceptions, boto3.exceptions):
        for name in dir(module):
            e = getattr(module, name)
            try:
                if issubclass(e, Exception):
                    setattr(hub.tool.boto3.exception, name, e)
            except TypeError:
                ...



import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

TREQ = {
    "present": {
        "require": [
            "aws.sns.topic.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    topic_arn: str,
    policy: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""

    Updates the topic's policy attribute. A topic can have a single policy, checks for changes in policy attribute and
    updates it if required.

    Args:
        hub:
        ctx:
        name(Text): The idem name for the topic_policy
        topic_arn(Text): The ARN of the topic for which the policy should be updated
        policy(Text): Topic policy, in json string format
        resource_id(Text, optional): Topic arn and 'policy' keyword separated with '-'



    Request Syntax:
        [topic-policy-name]:
          aws.sns.topic_policy.present:
          - name: 'string'
          - topic_arn: 'string'
          - policy: 'string'
          - resource_id: 'string'


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            topic-policy-name:
              aws.sns.topic_policy.present:
                - name: topic-policy-name
                - topic_arn: arn:aws:sns:eu-west-3:537227425989:test-topic
                - policy: '{"Version": "2012-10-17", "Id": "id-1", "Statement": [{"Sid":
                           "__default_statement_ID", "Effect": "Allow", "Principal": {"AWS": "*"}, "Action":
                           ["SNS:GetTopicAttributes", "SNS:SetTopicAttributes", "SNS:AddPermission", "SNS:RemovePermission",
                           "SNS:DeleteTopic", "SNS:Subscribe", "SNS:ListSubscriptionsByTopic", "SNS:Publish"],
                           "Resource": "arn:aws:sns:eu-west-3:537227425989:test-topic", "Condition": {"StringEquals":
                           {"AWS:SourceOwner": "537227425989"}}}]}'
                - resource_id: arn:aws:sns:eu-west-3:537227425989:test-topic-policy
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False
    plan_state = None

    # Standardise JSON string format
    policy = hub.tool.aws.sns.conversion_utils.standardise_json(policy)

    before = await hub.exec.boto3.client.sns.get_topic_attributes(
        ctx, TopicArn=topic_arn
    )
    if before:
        result[
            "old_state"
        ] = hub.tool.aws.sns.conversion_utils.convert_raw_topic_policy_to_present(
            raw_resource=before, idem_resource_name=name
        )

        plan_state = copy.deepcopy(result["old_state"])
        old_policy = plan_state.get("policy")

        if policy is not None:
            # Update policy attribute in topic
            if not hub.tool.aws.state_comparison_utils.is_json_identical(
                old_policy, policy
            ):
                if ctx.get("test", False):
                    plan_state["policy"] = policy
                    result["comment"] = (
                        f"Would update policy attribute of aws.sns.topic '{topic_arn}'",
                    )
                else:
                    update_attributes_ret = (
                        await hub.exec.boto3.client.sns.set_topic_attributes(
                            ctx,
                            TopicArn=topic_arn,
                            AttributeName="Policy",
                            AttributeValue=policy,
                        )
                    )
                    if not update_attributes_ret["result"]:
                        result["comment"] = update_attributes_ret["comment"]
                        result["result"] = False
                        return result
                    resource_updated = True
                    result["comment"] = (
                        f"Updated policy attribute of aws.sns.topic '{topic_arn}'",
                    )

            else:
                result["comment"] = (
                    f"No changes required for aws.sns.topic_policy '{name}'",
                )

    else:
        result["comment"] = (f"aws.sns.topic with '{topic_arn}' ARN, does not exist",)
        result["result"] = False

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif resource_updated:
        after = await hub.exec.boto3.client.sns.get_topic_attributes(
            ctx, TopicArn=topic_arn
        )
        result[
            "new_state"
        ] = hub.tool.aws.sns.conversion_utils.convert_raw_topic_policy_to_present(
            raw_resource=after, idem_resource_name=name
        )
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    r"""
    Deletes the current topic policy and replace with the default value. This action is idempotent, so deleting a topic's policy that does not exist
    does not result in an error.

    Args:
        name(Text): The idem name of the topic_policy.
        resource_id(Text): Topic arn and 'policy' keyword separated with '-'

    Request Syntax:
        [topic-policy-name]:
          aws.sns.topic_policy.absent:
          - name: 'string'
          - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test-topic-policy:
              aws.sns.topic_policy.absent:
              - name: test-topic-policy
              - resource_id: arn:aws:sns:eu-west-3:537227425989:test-topic-policy
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    topic_arn = resource_id.split("-policy")[0]
    before = await hub.exec.boto3.client.sns.get_topic_attributes(
        ctx, TopicArn=topic_arn
    )

    if not before:
        result["comment"] = (f"aws.sns.topic with '{topic_arn}' ARN does not exist",)

    else:
        result[
            "old_state"
        ] = hub.tool.aws.sns.conversion_utils.convert_raw_topic_policy_to_present(
            raw_resource=before, idem_resource_name=name
        )
        plan_state = copy.deepcopy(result["old_state"])
        default_policy = hub.tool.aws.sns.sns_utils.get_default_topic_policy(topic_arn)
        if ctx.get("test", False):
            plan_state["policy"] = default_policy
            result["new_state"] = plan_state
            result["comment"] = (f"Would delete aws.sns.topic_policy '{name}'",)
            return result

        # Deleting topic policy is not possible so replacing it with the default policy
        ret = await hub.exec.boto3.client.sns.set_topic_attributes(
            ctx,
            TopicArn=topic_arn,
            AttributeName="Policy",
            AttributeValue=default_policy,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result

        after = await hub.exec.boto3.client.sns.get_topic_attributes(
            ctx, TopicArn=topic_arn
        )
        result[
            "new_state"
        ] = hub.tool.aws.sns.conversion_utils.convert_raw_topic_policy_to_present(
            raw_resource=after, idem_resource_name=name
        )
        result["comment"] = (
            f"Deleted aws.sns.topic_policy for aws.sns.topic '{topic_arn}'",
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function
    Describes list of all the topic policy


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.sns.topic_policy
    """
    result = {}
    ret = await hub.exec.boto3.client.sns.list_topics(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe topic {ret['comment']}")
        return {}

    for topic in ret["ret"]["Topics"]:
        # Including fields to match the 'present' function parameters
        topic_arn = topic["TopicArn"]
        resource_id = topic_arn + "-policy"
        idem_resource_name = resource_id

        raw_resource = await hub.exec.boto3.client.sns.get_topic_attributes(
            ctx, TopicArn=topic_arn
        )

        resource_translated = (
            hub.tool.aws.sns.conversion_utils.convert_raw_topic_policy_to_present(
                raw_resource,
                idem_resource_name,
            )
        )
        result[resource_id] = {
            "aws.sns.topic_policy.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



"""
https://blog.gruntwork.io/authenticating-to-aws-with-the-credentials-file-d16c0fbcbf9e

$ aws configure
AWS Access Key ID: AKIAIOSFODNN7EXAMPLE
AWS Secret Access Key: wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY
Default region name [None]: us-west-2
Default output format [None]: json

AWS prompts you to enter your Access Key ID and Secret Access Key and stores them in ~/.aws/credentials:

[default]
aws_access_key_id=AKIAIOSFODNN7EXAMPLE
aws_secret_access_key=wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY

It also stores the other settings you entered in ~/.aws/config:

[default]
region=us-west-2
output=json
"""
import configparser
import copy
from typing import Any
from typing import Dict


async def gather(hub) -> Dict[str, Any]:
    """
    load profiles from unencrypted AWS credential files.

    Add "aws.iam" to the "acct.extras" section of your idem config file.

    Example:
    .. code-block:: yaml

        acct:
          extras:
            aws.iam:
              paths:
                - ~/.aws/credentials
                - /path/to/other/credential/file
              # Optional overrides
              region: us-east-1
    """
    sub_profiles = {}
    config = configparser.ConfigParser()

    try:
        ctx = copy.copy(hub.OPT.acct.extras.get("aws.iam", {}))
    except AttributeError:
        return sub_profiles

    credential_files = ctx.pop("paths", [])
    config.read(credential_files)

    for profile in set(config.sections()):
        sub_profiles[profile] = {name: value for name, value in config.items(profile)}
        sub_profiles[profile].update(ctx)

    return sub_profiles



from typing import Any
from typing import Dict
from typing import List


async def credentials(
    hub,
    ctx,
    role_arn: str,
    role_session_name: str,
    *,
    policy_arns: List[Dict[str, Any]] = None,
    policy: str = None,
    duration_seconds: int = None,
    tags: List[Dict[str, str]] = None,
    transitive_tag_keys: List[str] = None,
    external_id: str = None,
    serial_number: str = None,
    token_code: str = None,
    source_identity: str = None,
) -> Dict[str, Any]:
    """
    Returns a set of temporary security credentials that you can use to access Amazon Web Services resources that you might not normally have access to.
    These temporary credentials consist of an access key ID, a secret access key, and a security token.
    Typically, you use AssumeRole within your account or for cross-account access.
    For a comparison of AssumeRole with other API operations that produce temporary credentials,
    see Requesting Temporary Security Credentials and Comparing the Amazon Web Services STS API operations in the IAM User Guide .

    Examples:

    Call from the CLI:

    .. code-block:: bash

        $ idem exec aws.sts.assume_role.credentials <role_arn> <role_session_name>

    Call from code:

    .. code-block:: python

        async def my_func(hub, ctx, role_arn:str, role_session_name:str):
            await hub.exec.aws.sts.assume_role.credentials(ctx, role_arn, role_session_name)
    """
    ret = dict(result=True, ret={}, comment="")
    hub.log.debug(f"Assume role configuration set for ARN {role_arn}")

    config = {"RoleArn": role_arn, "RoleSessionName": role_session_name}
    if external_id is not None:
        config["ExternalId"] = external_id
    if policy_arns is not None:
        config["PolicyArns"] = policy_arns
    if policy is not None:
        config["Policy"] = policy
    if tags is not None:
        config["Tags"] = tags
    if transitive_tag_keys is not None:
        config["TransitiveTagKeys"] = transitive_tag_keys
    if duration_seconds is not None:
        config["DurationSeconds"] = duration_seconds
    if serial_number is not None:
        config["SerialNumber"] = serial_number
    if token_code is not None:
        config["TokenCode"] = token_code
    if source_identity is not None:
        config["SourceIdentity"] = source_identity

    assumed_role_object = await hub.exec.boto3.client.sts.assume_role(ctx, **config)
    ret["result"] = assumed_role_object.result
    ret["comment"] = assumed_role_object.comment
    if assumed_role_object.result:
        ret["ret"] = assumed_role_object.ret["Credentials"]
    return ret



from collections import OrderedDict
from datetime import timezone
from typing import Any
from typing import Dict

"""
Util functions to convert raw resource state from AWS EC2 to present input format.
"""
UTC_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


async def convert_raw_vpc_to_present(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("VpcId")
    resource_parameters = OrderedDict(
        {
            "InstanceTenancy": "instance_tenancy",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    if raw_resource.get("Tags") is not None:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )
    # The following code block is to make sure to only output the associated/associating cidr blocks and
    # dis-regard the disassociated cidr blocks.
    if raw_resource.get("CidrBlockAssociationSet"):
        ipv4_cidr_block_association_set = []
        for cidr_block in raw_resource.get("CidrBlockAssociationSet"):
            if cidr_block.get("CidrBlockState").get("State") in [
                "associated",
                "associating",
            ]:
                ipv4_cidr_block_association_set.append(cidr_block)
        resource_translated[
            "cidr_block_association_set"
        ] = ipv4_cidr_block_association_set
    if resource_id:
        enableDnsHostnames = await hub.exec.boto3.client.ec2.describe_vpc_attribute(
            ctx, Attribute="enableDnsHostnames", VpcId=resource_id
        )
        if enableDnsHostnames and enableDnsHostnames["result"] is True:
            resource_translated["enable_dns_hostnames"] = enableDnsHostnames["ret"][
                "EnableDnsHostnames"
            ]["Value"]
        else:
            # TODO - Need to handle the error efficiently.
            hub.log.warning(
                f"Failed on fetching enableDnsHostnames on vpc {resource_id} with error {enableDnsHostnames['comment']}."
            )
        enableDnsSupport = await hub.exec.boto3.client.ec2.describe_vpc_attribute(
            ctx, Attribute="enableDnsSupport", VpcId=resource_id
        )
        if enableDnsSupport and enableDnsSupport["result"] is True:
            resource_translated["enable_dns_support"] = enableDnsSupport["ret"][
                "EnableDnsSupport"
            ]["Value"]
        else:
            # TODO - Need to handle the error efficiently.
            hub.log.warning(
                f"Failed on fetching enableDnsSupport on vpc {resource_id} with error {enableDnsSupport['comment']}."
            )
    if raw_resource.get("Ipv6CidrBlockAssociationSet"):
        ipv6_cidr_block_association_set = []
        for cidr_block in raw_resource.get("Ipv6CidrBlockAssociationSet"):
            if cidr_block.get("Ipv6CidrBlockState").get("State") in [
                "associated",
                "associating",
            ]:
                if "NetworkBorderGroup" in cidr_block:
                    # Translate describe output to the correct present input format
                    ipv6_cidr_block_network_border_group = cidr_block.pop(
                        "NetworkBorderGroup"
                    )
                    cidr_block[
                        "Ipv6CidrBlockNetworkBorderGroup"
                    ] = ipv6_cidr_block_network_border_group
                ipv6_cidr_block_association_set.append(cidr_block)
        resource_translated[
            "ipv6_cidr_block_association_set"
        ] = ipv6_cidr_block_association_set
    return resource_translated


def convert_raw_subnet_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("SubnetId")
    resource_parameters = OrderedDict(
        {
            "VpcId": "vpc_id",
            "CidrBlock": "cidr_block",
            "AvailabilityZone": "availability_zone",
            "OutpostArn": "outpost_arn",
            "MapPublicIpOnLaunch": "map_public_ip_on_launch",
            "AssignIpv6AddressOnCreation": "assign_ipv6_address_on_creation",
            "MapCustomerOwnedIpOnLaunch": "map_customer_owned_ip_on_launch",
            "CustomerOwnedIpv4Pool": "customer_owned_ipv4_pool",
            "EnableDns64": "enable_dns_64",
            "PrivateDnsNameOptionsOnLaunch": "private_dns_name_options_on_launch",
            "EnableLniAtDeviceIndex": "enable_lni_at_device_index",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    if "Tags" in raw_resource:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )
    if (not raw_resource.get("AvailabilityZone")) and raw_resource.get(
        "AvailabilityZoneId"
    ):
        # Only populate availability_zone_id field when availability_zone doesn't exist
        resource_translated["availability_zone_id"] = raw_resource.get(
            "AvailabilityZoneId"
        )
    if raw_resource.get("Ipv6CidrBlockAssociationSet"):
        ipv6_cidr_block_association_set = (
            hub.tool.aws.network_utils.get_associated_ipv6_cidr_blocks(
                raw_resource.get("Ipv6CidrBlockAssociationSet")
            )
        )
        # We should only output the associated ipv6 cidr block, and theoretically there should only be one,
        # since AWS only supports one ipv6 cidr block association on a subnet
        if ipv6_cidr_block_association_set:
            resource_translated["ipv6_cidr_block"] = ipv6_cidr_block_association_set[
                0
            ].get("Ipv6CidrBlock")
    return resource_translated


def convert_raw_sg_rule_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("SecurityGroupRuleId")
    resource_parameters = OrderedDict(
        {
            "GroupId": "group_id",
            "IsEgress": "is_egress",
            "IpProtocol": "ip_protocol",
            "FromPort": "from_port",
            "ToPort": "to_port",
            "CidrIpv4": "cidr_ipv4",
            "CidrIpv6": "cidr_ipv6",
            "PrefixListId": "prefix_list_id",
            "Description": "description",
            "ReferencedGroupInfo": "referenced_group_info",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    if "Tags" in raw_resource:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )
    return resource_translated


def convert_raw_transit_gateway_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("TransitGatewayId")
    resource_parameters = OrderedDict(
        {
            "State": "state",
            "Description": "description",
            "Options": "options",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    if "Tags" in raw_resource:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )
    return resource_translated


def convert_raw_key_pair_to_present(
    hub, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Convert the aws key pair response to a common format

    @param raw_resource: describe key pair response
    @param hub: required for functions in hub

    @return:
        A dictionary key pair's properties
    """
    if not raw_resource:
        return {}

    resource_parameters = OrderedDict(
        {
            "KeyPairId": "resource_id",
            "PublicKey": "public_key",
        }
    )
    resource_translated = {"name": raw_resource.get("KeyName")}
    for raw_parameter, present_parameter in resource_parameters.items():
        if raw_resource.get(raw_parameter):
            resource_translated[present_parameter] = raw_resource.get(raw_parameter)
    if raw_resource.get("Tags") is not None:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )
    return resource_translated


def convert_raw_placement_group_to_present(
    hub, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Convert the aws placement_group response to a common format

    Args:
        hub: required for functions in hub
        raw_resource: The aws response to convert

    Returns: Valid idem state for placement_group of type Dict['string', Any]
    """
    if not raw_resource:
        return {}

    resource_parameters = OrderedDict(
        {
            "State": "state",
            "Strategy": "strategy",
            "PartitionCount": "partition_count",
            "GroupId": "group_id",
            "GroupName": "group_name",
            "SpreadLevel": "spread_level",
        }
    )
    resource_translated = {"resource_id": raw_resource.get("GroupName", None)}
    for raw_parameter, present_parameter in resource_parameters.items():
        if raw_resource.get(raw_parameter):
            resource_translated[present_parameter] = raw_resource.get(raw_parameter)
    if raw_resource.get("Tags") is not None:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )
    return resource_translated


def convert_raw_flow_log_to_present(hub, flow_log: Dict):
    """
    Convert the aws flow log response to a common format

    Args:
        hub: required for functions in hub
        flow_log: describe flow log response

    Returns:
        A dictionary of flow log
    """
    describe_parameters = OrderedDict(
        {
            "DeliverLogsPermissionArn": "iam_role",
            "LogGroupName": "log_group_name",
            "TrafficType": "traffic_type",
            "LogDestinationType": "log_destination_type",
            "LogDestination": "log_destination",
            "LogFormat": "log_format",
            "MaxAggregationInterval": "max_aggregation_interval",
        }
    )
    translated_flow_log = {}
    flow_log_id = flow_log.get("FlowLogId")
    resource_id = flow_log.get("ResourceId")
    resource_ids = [resource_id]
    resource_type = resource_id.split("-")[0]
    translated_flow_log["resource_ids"] = resource_ids
    if resource_type == "vpc":
        resource_type = "VPC"
    elif resource_type == "subnet":
        resource_type = "Subnet"
    else:
        resource_type = "NetworkInterface"
    translated_flow_log["resource_type"] = resource_type
    translated_flow_log["resource_id"] = flow_log_id

    for parameter_old_key, parameter_new_key in describe_parameters.items():
        if flow_log.get(parameter_old_key) is not None:
            translated_flow_log[parameter_new_key] = flow_log.get(parameter_old_key)
    if flow_log.get("Tags") is not None:
        translated_flow_log["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            flow_log.get("Tags")
        )
    return translated_flow_log


def convert_raw_route_table_to_present(
    hub, raw_resource: Dict, idem_resource_name: str = None
) -> Dict[str, Any]:
    describe_parameters = OrderedDict(
        {"VpcId": "vpc_id", "PropagatingVgws": "propagating_vgws", "Tags": "tags"}
    )
    if not raw_resource:
        return {}
    resource_id = raw_resource.get("RouteTableId")
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_old_key, parameter_new_key in describe_parameters.items():
        if raw_resource.get(parameter_old_key) is not None:
            resource_translated[parameter_new_key] = raw_resource.get(parameter_old_key)
    routes_to_add = []
    if raw_resource.get("Routes"):
        for route_to_add in raw_resource.get("Routes"):
            if route_to_add.get("State") == "active":
                routes_to_add.append(route_to_add)
        resource_translated["routes"] = routes_to_add
    if raw_resource.get("Tags") is not None:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )
    return resource_translated


def convert_raw_route_table_association_to_present(
    hub, resource: Dict, idem_resource_name: str = None
) -> Dict[str, Any]:
    if not resource:
        return {}
    resource_id = resource.get("RouteTableAssociationId")
    resource_parameters = OrderedDict(
        {
            "RouteTableId": "route_table_id",
            "GatewayId": "gateway_id",
            "SubnetId": "subnet_id",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in resource:
            resource_translated[parameter_present] = resource.get(parameter_raw)
    return resource_translated


def convert_raw_tg_vpc_attachment(
    hub, raw_resource: Dict, resource_name: str = None
) -> Dict[str, any]:
    resource_id = raw_resource.get("TransitGatewayAttachmentId")
    translated_resource = {"name": resource_name, "resource_id": resource_id}

    resource_parameters_transit_gateway = OrderedDict(
        {
            "TransitGatewayId": "transit_gateway",
            "VpcId": "vpc_id",
            "State": "state",
            "SubnetIds": "subnet_ids",
            "Options": "options",
        }
    )

    for camel_case_key, snake_case_key in resource_parameters_transit_gateway.items():
        if raw_resource.get(camel_case_key):
            translated_resource[snake_case_key] = raw_resource.get(camel_case_key)
    if raw_resource.get("Tags") is not None:
        translated_resource["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )
    return translated_resource


async def convert_raw_dhcp_to_present(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("DhcpOptionsId")
    resource_parameters = OrderedDict({"Tags": "tags"})
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    result = dict(comment=(), result=True, ret=None)
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    if raw_resource.get("DhcpConfigurations"):
        dhcp_configs = []
        for dhcp_conf in raw_resource.get("DhcpConfigurations"):
            dhcp_config_values_list = []
            for dhcp in dhcp_conf.get("Values"):
                if "Value" in dhcp:
                    dhcp_config_values_list.append(dhcp.get("Value"))
            dhcp_config = {
                "Key": dhcp_conf.get("Key"),
                "Values": dhcp_config_values_list,
            }
            dhcp_configs.append(dhcp_config)
        resource_translated["dhcp_configurations"] = dhcp_configs
    vpc_list = []
    vpc_ret = await hub.exec.boto3.client.ec2.describe_vpcs(ctx)
    if not vpc_ret["result"]:
        result["comment"] = vpc_ret["comment"]
        result["result"] = vpc_ret["result"]
    else:
        for vpc in vpc_ret["ret"]["Vpcs"]:
            if vpc["DhcpOptionsId"] == resource_id:
                vpc_list.append(vpc["VpcId"])
    if vpc_list:
        resource_translated["vpc_id"] = vpc_list

    if raw_resource.get("Tags") is not None:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )
    result["ret"] = resource_translated
    return result


def convert_raw_dhcp_association_to_present(
    hub, dhcp_id: str, idem_resource_name: str = None, vpc_id: str = None
) -> Dict[str, Any]:
    """
    Convert the AWS EC2 DHCP Options Association response to a common format

    Args:
        hub: required for functions in hub
        dhcp_id: ID of DHCP Options.
        idem_resource_name: name of the resource
        vpc_id: ID of VPC to which DHCP options need to be attached.

    Returns:
        A dictionary of EC2 DHCP options.
    """
    resource_translated = {
        "name": idem_resource_name,
        "dhcp_id": dhcp_id,
        "resource_id": f"{dhcp_id}-{vpc_id}",
    }
    if vpc_id:
        resource_translated["vpc_id"] = vpc_id
    return resource_translated


def convert_raw_sg_to_present(hub, security_group: Dict):
    """
    Convert the aws security group response to a common format

    Args:
        hub: required for functions in hub
        security_group: describe security group response

    Returns:
        A dictionary of sg group
    """
    translated_security_group = {}
    describe_parameters = OrderedDict(
        {
            "GroupId": "resource_id",
            "GroupName": "name",
            "VpcId": "vpc_id",
            "Description": "description",
        }
    )
    for parameter_old_key, parameter_new_key in describe_parameters.items():
        if security_group.get(parameter_old_key) is not None:
            translated_security_group[parameter_new_key] = security_group.get(
                parameter_old_key
            )
    if "Tags" in security_group:
        translated_security_group[
            "tags"
        ] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(security_group.get("Tags"))

    return translated_security_group


def convert_raw_internet_gateway_to_present(
    hub, resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource.get("InternetGatewayId"),
    }
    if resource.get("Tags"):
        resource_translated["tags"] = resource.get("Tags")
    if resource.get("Attachments"):
        resource_translated["vpc_id"] = [resource.get("Attachments")[0].get("VpcId")]
        resource_translated["attachments"] = resource.get("Attachments")
    if resource.get("Tags") is not None:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            resource.get("Tags")
        )
    return resource_translated


def convert_raw_nat_gateway_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("NatGatewayId")
    describe_parameters = OrderedDict(
        {
            "SubnetId": "subnet_id",
            "ConnectivityType": "connectivity_type",
            "State": "state",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in describe_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    if raw_resource.get("NatGatewayAddresses"):
        for nat_gateway_address in raw_resource.get("NatGatewayAddresses"):
            if "AllocationId" in nat_gateway_address:
                resource_translated["allocation_id"] = nat_gateway_address.get(
                    "AllocationId"
                )
                break
    if "Tags" in raw_resource:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )
    return resource_translated


def convert_raw_elastic_ip_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    """
    Given an object state from aws, this function creates a translated resource object in response.

    Args:
        hub: required for functions in hub
        raw_resource (Dict): The dictionary object from where the raw state of resource needs to be translated.
        idem_resource_name (Text): The name of the idem resource

    Returns: Dict[str, Any]
    """
    elastic_ip = raw_resource
    resource_id = elastic_ip.get("PublicIp")
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    describe_parameters = OrderedDict(
        {
            "AllocationId": "allocation_id",
            "AssociationId": "association_id",
            "Domain": "domain",
            "InstanceId": "instance_id",
            "NetworkBorderGroup": "network_border_group",
            "PublicIpv4Pool": "public_ipv4_pool",
            "Tags": "tags",
            "PublicIp": "resource_id",
            "CustomerOwnedIpv4Pool": "customer_owned_ipv4_pool",
        }
    )
    for parameter_raw, parameter_present in describe_parameters.items():
        if elastic_ip.get(parameter_raw) is not None:
            resource_translated[parameter_present] = elastic_ip.get(parameter_raw)
    if raw_resource.get("Tags") is not None:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )
    return resource_translated


def convert_raw_vpc_endpoint_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    """
    Convert AWS returned data structure to correct idem vpc-endpoint present state

     Args:
        hub: required for functions in hub
        raw_resource: The aws response to convert

    Returns: Valid idem state for vpc-endpoint of type Dict['string', Any]
    """
    hub.log.debug(
        f"Converting raw aws.ec2.vpc_endpoint: {raw_resource} to valid present state"
    )
    describe_parameters = OrderedDict(
        {
            "VpcEndpointType": "vpc_endpoint_type",
            "VpcId": "vpc_id",
            "ServiceName": "service_name",
            "PolicyDocument": "policy_document",
            "RouteTableIds": "route_table_ids",
            "SubnetIds": "subnet_ids",
            "SecurityGroupIds": "security_group_ids",
            "PrivateDnsEnabled": "private_dns_enabled",
            "State": "state",
        }
    )
    resource_id = raw_resource.get("VpcEndpointId")
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in describe_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    if raw_resource.get("Tags") is not None:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )

    return resource_translated


def convert_raw_ami_to_present(hub, raw_resource: Dict[str, Any]) -> Dict[str, Any]:
    """
    Util function to convert raw resource state from AWS AMI to present input format.
    """
    resource_id = raw_resource.get("ImageId")
    resource_parameters = OrderedDict(
        {
            "Name": "name",
            "CreationDate": "creation_date",
            "ImageLocation": "image_location",
            "Architecture": "architecture",
            "KernelId": "kernel_id",
            "BlockDeviceMappings": "block_device_mappings",
            "Description": "description",
            "EnaSupport": "ena_support",
            "BillingProducts": "billing_products",
            "RamdiskId": "ramdisk_id",
            "RootDeviceName": "root_device_name",
            "SriovNetSupport": "sriov_net_support",
            "VirtualizationType": "virtualization_type",
            "BootMode": "boot_mode",
        }
    )
    resource_translated = {"resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    if "Tags" in raw_resource:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )
    return resource_translated


def convert_raw_vpc_peering_connection_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str
) -> Dict[str, Any]:
    if not raw_resource:
        return {}

    resource_translated = {
        "resource_id": raw_resource["VpcPeeringConnectionId"],
        "name": idem_resource_name,
        "peer_owner_id": raw_resource["AccepterVpcInfo"]["OwnerId"],
        "peer_vpc_id": raw_resource["AccepterVpcInfo"]["VpcId"],
        "vpc_id": raw_resource["RequesterVpcInfo"]["VpcId"],
        "peer_region": raw_resource["AccepterVpcInfo"]["Region"],
        "tags": hub.tool.aws.tag_utils.convert_tag_list_to_dict(raw_resource["Tags"]),
        "status": raw_resource["Status"]["Code"],
    }

    return resource_translated


def convert_raw_availability_zone_to_snake_case(
    hub, ctx, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Util function to convert raw resource state from AWS Availability Zone to a snake case format.
    """
    if not raw_resource:
        return {}

    resource_parameters = OrderedDict(
        {
            "GroupName": "group_name",
            "NetworkBorderGroup": "network_border_group",
            "OptInStatus": "opt_in_status",
            "ParentZoneId": "parent_zone_id",
            "ParentZoneName": "parent_zone_name",
            "RegionName": "region_name",
            "State": "state",
            "ZoneName": "zone_name",
            "ZoneType": "zone_type",
        }
    )
    resource_translated = {"resource_id": raw_resource["ZoneId"]}
    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


async def convert_raw_snapshot_to_present(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str
) -> Dict[str, Any]:
    """
    Util function to convert raw resource state from AWS snapshot to present input format
    """
    resource_id = raw_resource.get("SnapshotId")
    resource_parameters = OrderedDict(
        {
            "Description": "description",
            "OutpostArn": "outpost_arn",
            "Tags": "tags",
            "VolumeId": "volume_id",
        }
    )
    resource_translated = {"resource_id": resource_id}
    resource_translated["name"] = idem_resource_name
    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    """
        Expansion of the volume associated with the snapshot, if present. This is due to the one-way
        relationship with volume once the snapshot is created, to allow the user access to up to date
        values of the volume if it has not been removed. This also fills a feature gap with terraform
        which provides some of this expansion as part of its snapshot support.
    """
    if resource_translated["volume_id"] is not None:
        volume = await hub.exec.aws.ec2.volume.get(
            ctx=ctx,
            resource_id=resource_translated["volume_id"],
            name=resource_translated["name"],
        )

    if volume["result"] and volume["ret"]:
        translated_volume = {}
        for prop in volume["ret"]:
            if prop != "tags":
                translated_volume[prop] = volume["ret"].get(prop)
        resource_translated["associated_volume"] = translated_volume

    if "StartTime" in raw_resource:
        local_time = raw_resource.get("StartTime").astimezone()
        start_time = local_time.astimezone(timezone.utc).strftime(UTC_FORMAT)
        resource_translated["start_time"] = start_time

    if "Tags" in raw_resource:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )

    return resource_translated


def convert_raw_volume_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str
) -> Dict[str, Any]:
    """
    Util function to convert raw resource state from AWS Volume to present input format
    """
    resource_id = raw_resource.get("VolumeId")
    resource_parameters = OrderedDict(
        {
            "AvailabilityZone": "availability_zone",
            "Size": "size",
            "SnapshotId": "snapshot_id",
            "State": "state",
            "Iops": "iops",
            "Tags": "tags",
            "VolumeType": "volume_type",
            "MultiAttachEnabled": "multi_attach_enabled",
            "Throughput": "throughput",
            "Encrypted": "encrypted",
        }
    )
    resource_translated = {"resource_id": resource_id}
    resource_translated["name"] = idem_resource_name
    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    if "Tags" in raw_resource:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )
    return resource_translated


def convert_raw_vpc_peering_connection_options_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str
) -> Dict[str, Any]:
    if not raw_resource:
        return {}

    resource_translated = {
        "resource_id": raw_resource["VpcPeeringConnectionId"],
        "name": idem_resource_name,
        "peer_allow_remote_vpc_dns_resolution": raw_resource["AccepterVpcInfo"][
            "PeeringOptions"
        ]["AllowDnsResolutionFromRemoteVpc"],
        "peer_allow_classic_link_to_remote_vpc": raw_resource["AccepterVpcInfo"][
            "PeeringOptions"
        ]["AllowEgressFromLocalClassicLinkToRemoteVpc"],
        "peer_allow_vpc_to_remote_classic_link": raw_resource["AccepterVpcInfo"][
            "PeeringOptions"
        ]["AllowEgressFromLocalVpcToRemoteClassicLink"],
        "allow_remote_vpc_dns_resolution": raw_resource["RequesterVpcInfo"][
            "PeeringOptions"
        ]["AllowDnsResolutionFromRemoteVpc"],
        "allow_classic_link_to_remote_vpc": raw_resource["RequesterVpcInfo"][
            "PeeringOptions"
        ]["AllowEgressFromLocalClassicLinkToRemoteVpc"],
        "allow_vpc_to_remote_classic_link": raw_resource["RequesterVpcInfo"][
            "PeeringOptions"
        ]["AllowEgressFromLocalVpcToRemoteClassicLink"],
    }

    return resource_translated



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


def convert_raw_db_subnet_group_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    tags: Dict[str, Any] = None,
) -> Dict[str, Any]:
    """
    Convert AWS returned data structure to correct idem db_subnet_group present state

     Args:
        hub: required for functions in hub
        raw_resource (Dict): The aws response to convert
        idem_resource_name (Text, Optional): The idem name of the resource
        tags (Dict, Optional): The tags of the resource. Defaults to None.

    Returns: Valid idem state for db_subnet_group of type Dict['string', Any]
    """
    resource_id = raw_resource["DBSubnetGroupName"]
    raw_resource["Tags"] = tags
    resource_parameters = OrderedDict(
        {
            "DBSubnetGroupDescription": "db_subnet_group_description",
            "Subnets": "subnet_ids",
            "Tags": "tags",
            "DBSubnetGroupArn": "db_subnet_group_arn",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            if parameter_raw == "Subnets":
                resource_translated[parameter_present] = [
                    subnet["SubnetIdentifier"] for subnet in raw_resource.get("Subnets")
                ]
            else:
                resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


async def search_raw(
    hub, ctx, filters: List = None, DBSubnetGroupName: str = None
) -> Dict:
    """

    Args:
        filters(list, optional):
        DBSubnetGroupName(string, optional): Name of DB Subnet Group.

    Returns:
        {"result": True|False, "comment": A message List, "ret": Dict}
    """
    result = dict(comment=[], ret=None, result=True)
    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    if not syntax_validation["result"]:
        result["result"] = False
        result["comment"] = list(syntax_validation["comment"])
        return result
    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )
    ret = await hub.exec.boto3.client.docdb.describe_db_subnet_groups(
        ctx,
        Filters=boto3_filter,
        DBSubnetGroupName=DBSubnetGroupName,
    )
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result



"""
Autogenerated state module using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.create_subnet
hub.exec.boto3.client.ec2.delete_subnet
hub.exec.boto3.client.ec2.describe_subnets
resource = hub.tool.boto3.resource.create(ctx, "ec2", "Subnet", name)
hub.tool.boto3.resource.exec(resource, create_instances, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_network_interface, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_tags, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

TREQ = {
    "absent": {
        "require": [
            "aws.ec2.nat_gateway.absent",
            "aws.ec2.elastic_ip.absent",
            "aws.ec2.route_table.absent",
            "aws.eks.cluster.absent",
        ],
    },
    "present": {
        "require": [
            "aws.ec2.vpc.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    vpc_id: str,
    cidr_block: str,
    resource_id: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
    availability_zone: str = None,
    availability_zone_id: str = None,
    ipv6_cidr_block: str = None,
    outpost_arn: str = None,
    map_public_ip_on_launch: bool = None,
    assign_ipv6_address_on_creation: bool = None,
    map_customer_owned_ip_on_launch: bool = None,
    customer_owned_ipv4_pool: str = None,
    enable_dns_64: bool = None,
    private_dns_name_options_on_launch: Dict[str, Any] = None,
    enable_lni_at_device_index: int = None,
    disable_lni_at_device_index: bool = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a subnet in a specified VPC. You must specify an IPv4 CIDR block for the subnet. After you create a
    subnet, you can't change its CIDR block. The allowed block size is between a /16 netmask (65,536 IP addresses)
    and /28 netmask (16 IP addresses). The CIDR block must not overlap with the CIDR block of an existing subnet in
    the VPC. If you've associated an IPv6 CIDR block with your VPC, you can create a subnet with an IPv6 CIDR block
    that uses a /64 prefix length.   Amazon Web Services reserves both the first four and the last IPv4 address in
    each subnet's CIDR block. They're not available for use.  If you add more than one subnet to a VPC, they're set
    up in a star topology with a logical router in the middle. When you stop an instance in a subnet, it retains its
    private IPv4 address. It's therefore possible to have a subnet with no running instances (they're all stopped),
    but no remaining IP addresses available. For more information about subnets, see Your VPC and subnets in the
    Amazon Virtual Private Cloud User Guide.

    Args:
        name(Text): An Idem name of the resource.
        vpc_id(Text): ID of the VPC.
        cidr_block(Text): The IPv4 network range for the subnet, in CIDR notation. For example, 10.0.0.0/24. We modify the
            specified CIDR block to its canonical form; for example, if you specify 100.68.0.18/18, we
            modify it to 100.68.0.0/18.
        resource_id(Text, optional): AWS Subnet ID
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the subnet.
            Each tag consists of a key name and an associated value. Defaults to None.
            * (Key, optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * (Value, optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.
        availability_zone(Text, optional): The Availability Zone or Local Zone for the subnet. Default: Amazon Web Services selects one for
            you. If you create more than one subnet in your VPC, we do not necessarily select a different
            zone for each subnet. To create a subnet in a Local Zone, set this value to the Local Zone ID,
            for example us-west-2-lax-1a. For information about the Regions that support Local Zones, see
            Available Regions in the Amazon Elastic Compute Cloud User Guide. To create a subnet in an
            Outpost, set this value to the Availability Zone for the Outpost and specify the Outpost ARN. Defaults to None.
        availability_zone_id(Text, optional): The AZ ID or the Local Zone ID of the subnet. Defaults to None.
        ipv6_cidr_block(Text, optional): The IPv6 network range for the subnet, in CIDR notation. The subnet size must use a /64 prefix
            length. Defaults to None.
        outpost_arn(Text, optional): The Amazon Resource Name (ARN) of the Outpost. If you specify an Outpost ARN, you must also
            specify the Availability Zone of the Outpost subnet. Defaults to None.
        map_public_ip_on_launch (boolean, optional): Indicates whether instances launched in this subnet receive a public IPv4 address.
        assign_ipv6_address_on_creation (boolean, optional): Specify true to indicate that network interfaces created in the specified subnet should be assigned an IPv6 address.
            This includes a network interface that's created when launching an instance into the subnet (the instance therefore receives an IPv6 address).
        map_customer_owned_ip_on_launch (boolean, optional): Specify true to indicate that network interfaces attached to instances created in the specified subnet should be assigned a customer-owned IPv4 address.
        customer_owned_ipv4_pool (Text, optional): The customer-owned IPv4 address pool associated with the subnet. You must set this value when you specify true for MapCustomerOwnedIpOnLaunch .
        enable_dns_64 (boolean, optional): Indicates whether DNS queries made to the Amazon-provided DNS Resolver in this subnet should return synthetic IPv6 addresses for IPv4-only destinations.
        enable_lni_at_device_index (int, optional): Indicates the device position for local network interfaces in this subnet.
        private_dns_name_options_on_launch (Dict) -- The type of hostnames to assign to instances in the subnet at launch. An instance hostname is based on the IPv4 address or ID of the instance.
            HostnameType (Text) -- The type of hostname for EC2 instances. For IPv4 only subnets, an instance DNS name must be based on the instance IPv4 address.
                For IPv6 only subnets, an instance DNS name must be based on the instance ID. For dual-stack subnets, you can specify whether DNS names use the instance IPv4 address or the instance ID.
            EnableResourceNameDnsARecord (boolean) -- Indicates whether to respond to DNS queries for instance hostnames with DNS A records.
            EnableResourceNameDnsAAAARecord (boolean) -- Indicates whether to respond to DNS queries for instance hostname with DNS AAAA records.
        disable_lni_at_device_index (boolean): Specify true to indicate that local network interfaces at the current position should be disabled.

    Request Syntax:
        [subnet-resource-name]:
          aws.ec2.subnet.present:
          - resource_id: 'string'
          - cidr_block: 'string'
          - ipv6_cidr_block: 'string'
          - vpc_id: 'string'
          - availability_zone: 'string'
          - availability_zone_id: 'string'
          - outpost_arn: 'string'
          - map_public_ip_on_launch: bool
          - assign_ipv6_address_on_creation: bool
          - map_customer_owned_ip_on_launch: bool
          - enable_dns_64: bool
          - private_dns_name_options_on_launch:
              EnableResourceNameDnsAAAARecord: bool
              EnableResourceNameDnsARecord: bool
              HostnameType: 'string'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            my-subnet:
              aws.ec2.subnet.present:
                - vpc_id: vpc-07123af5a5zwqcc0
                - cidr_block: 10.10.10.0/28
                - availability_zone: eu-west-2c
                - tags:
                  - Key: Name
                    Value: Idem-test-subnet
                - ipv6_cidr_block: 2a05:d01c:74f:7200::/64
                - map_public_ip_on_launch: true
                - assign_ipv6_address_on_creation: false
                - map_customer_owned_ip_on_launch: false
                - enable_dns_64: false
                - private_dns_name_options_on_launch:
                    EnableResourceNameDnsAAAARecord: false
                    EnableResourceNameDnsARecord: false
                    HostnameType: ip-name


    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False
    before = None

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if resource_id:
        before = await hub.exec.aws.ec2.subnet.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        # Update ipv6 cidr block
        ipv6_cidr_block_association_set = (
            hub.tool.aws.network_utils.get_associated_ipv6_cidr_blocks(
                before["ret"].get("Ipv6CidrBlockAssociationSet")
            )
        )
        if ipv6_cidr_block != (
            None
            if not ipv6_cidr_block_association_set
            else ipv6_cidr_block_association_set[0]
        ):
            update_ret = await hub.tool.aws.ec2.subnet.update_ipv6_cidr_blocks(
                ctx=ctx,
                subnet_id=before["ret"]["resource_id"],
                old_ipv6_cidr_block=None
                if not ipv6_cidr_block_association_set
                else ipv6_cidr_block_association_set[0],
                new_ipv6_cidr_block={"Ipv6CidrBlock": ipv6_cidr_block}
                if ipv6_cidr_block
                else None,
            )
            result["result"] = result["result"] and update_ret["result"]
            result["comment"] = update_ret["comment"]
            resource_updated = True
            if ctx.get("test", False) and update_ret["ret"] is not None:
                plan_state["ipv6_cidr_block"] = update_ret["ret"].get("ipv6_cidr_block")
        if tags is not None and tags != result["old_state"].get("tags"):
            # Update tags
            update_ret = await hub.exec.aws.ec2.tag.update_tags(
                ctx=ctx,
                resource_id=before["ret"]["resource_id"],
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = result["result"] and update_ret["result"]
            resource_updated = resource_updated or update_ret["result"]
            if ctx.get("test", False) and update_ret["result"]:
                plan_state["tags"] = update_ret["ret"]
        if resource_updated:
            if ctx.get("test", False):
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.ec2.subnet", name=name
                )
            else:
                result["comment"] += hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.ec2.subnet", name=name
                )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "availability_zone": availability_zone,
                    "availability_zone_id": availability_zone_id,
                    "cidr_block": cidr_block,
                    "ipv6_cidr_block": ipv6_cidr_block,
                    "outpost_arn": outpost_arn,
                    "vpc_id": vpc_id,
                    "tags": tags,
                    "map_public_ip_on_launch": map_public_ip_on_launch,
                    "assign_ipv6_address_on_creation": assign_ipv6_address_on_creation,
                    "map_customer_owned_ip_on_launch": map_customer_owned_ip_on_launch,
                    "customer_owned_ipv4_pool": customer_owned_ipv4_pool,
                    "enable_dns_64": enable_dns_64,
                    "private_dns_name_options_on_launch": private_dns_name_options_on_launch,
                    "enable_lni_at_device_index": enable_lni_at_device_index,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ec2.subnet", name=name
            )
            return result
        ret = await hub.exec.boto3.client.ec2.create_subnet(
            ctx,
            **{
                "TagSpecifications": [
                    {
                        "ResourceType": "subnet",
                        "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
                    }
                ]
                if tags
                else None,
                "AvailabilityZone": availability_zone,
                "AvailabilityZoneId": availability_zone_id,
                "Ipv6CidrBlock": ipv6_cidr_block,
                "OutpostArn": outpost_arn,
                "VpcId": vpc_id,
                "CidrBlock": cidr_block,
            },
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.ec2.subnet", name=name
        )
        result[
            "new_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_subnet_to_present(
            raw_resource=ret["ret"]["Subnet"], idem_resource_name=name
        )
        resource_id = result["new_state"]["resource_id"]

    if not before:
        new_res = result["new_state"]
    else:
        new_res = before["ret"]
    update_ret = await hub.tool.aws.ec2.subnet.update_subnet_attributes(
        ctx,
        before=new_res,
        resource_id=resource_id,
        map_public_ip_on_launch=map_public_ip_on_launch,
        assign_ipv6_address_on_creation=assign_ipv6_address_on_creation,
        map_customer_owned_ip_on_launch=map_customer_owned_ip_on_launch,
        customer_owned_ipv4_pool=customer_owned_ipv4_pool,
        enable_dns_64=enable_dns_64,
        private_dns_name_options_on_launch=private_dns_name_options_on_launch,
        enable_lni_at_device_index=enable_lni_at_device_index,
        disable_lni_at_device_index=disable_lni_at_device_index,
    )
    result["comment"] = result["comment"] + update_ret["comment"]
    result["result"] = result["result"] and update_ret["result"]
    resource_updated = resource_updated or bool(update_ret["ret"])
    if update_ret["ret"] and ctx.get("test", False):
        plan_state.update(update_ret["ret"])

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not before) or resource_updated:
        after = await hub.exec.aws.ec2.subnet.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not after["result"] or not after["ret"]:
            result["result"] = False
            result["comment"] = after["comment"]
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified subnet. You must terminate all running instances in the subnet before you can delete the
    subnet.

    Args:
        name(Text): The Idem name of the subnet.
        resource_id(Text, optional): AWS Subnet ID. Idem automatically considers this resource being absent if
         this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.ec2.subnet.absent:
                - name: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.subnet", name=name
        )
        return result

    before = await hub.exec.boto3.client.ec2.describe_subnets(
        ctx, SubnetIds=[resource_id]
    )
    if not before["result"]:
        if "InvalidSubnetID.NotFound" in str(before["comment"]):
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.ec2.subnet", name=name
            )
        else:
            result["comment"] = before["comment"]
            result["result"] = False
        return result
    if not before["ret"].get("Subnets"):
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.subnet", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_subnet_to_present(
            raw_resource=before["ret"]["Subnets"][0], idem_resource_name=name
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.ec2.subnet", name=name
        )
        return result
    else:
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_subnet_to_present(
            raw_resource=before["ret"]["Subnets"][0], idem_resource_name=name
        )
        try:
            ret = await hub.exec.boto3.client.ec2.delete_subnet(
                ctx, SubnetId=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.ec2.subnet", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_subnets(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe Subnets {ret['comment']}")
        return {}

    for subnet in ret["ret"]["Subnets"]:
        subnet_id = subnet.get("SubnetId")
        resource_translated = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_subnet_to_present(
                raw_resource=subnet, idem_resource_name=subnet_id
            )
        )
        result[subnet_id] = {
            "aws.ec2.subnet.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result


async def search(
    hub,
    ctx,
    name,
    resource_id: str = None,
    availability_zone: str = None,
    availability_zone_id: str = None,
    cidr_block: str = None,
    default_for_az: str = None,
    filters: List = None,
    ipv6_cidr_block: str = None,
    status: str = None,
    vpc_id: str = None,
    tags: List[Dict[str, Any]] or Dict[str, Any] = None,
):
    """
    Use an un-managed subnet as a data-source. Supply one of the inputs as the filter. This function has been deprecated.
    Please use exec.run with aws.ec2.subnet.get instead.

    Args:
        name(string): The name of the Idem state.
        resource_id(string, optional): AWS subnet id to identify the resource.
        availability_zone(string, optional): The Availability Zone for the subnet.
        availability_zone_id(string, optional): The ID of the Availability Zone for the subnet.
        cidr_block(string, optional): The IPv4 CIDR block of the subnet. The CIDR block you specify must exactly match
         the subnet's CIDR block for information to be returned for the subnet.
        default_for_az(string, optional): Indicate whether the subnet is the default subnet in the Availability Zone.
        filters(List, optional): One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
         https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_subnets
        ipv6_cidr_block(string, optional): An IPv6 CIDR block associated with the subnet.
        status(string, optional): The state of the subnet (pending | available ).
        vpc_id(string, optional): The ID of the VPC for the subnet.
        tags(List or Dict, optional): The list or dict of tags to filter by.

    """
    hub.log.warning(
        f"aws.ec2.subnet.search '{name}' state has been deprecated. Please use exec.run with aws.ec2.subnet.get instead."
    )

    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)
    ret = await hub.exec.aws.ec2.subnet.get(
        ctx=ctx,
        name=name,
        resource_id=resource_id,
        availability_zone=availability_zone,
        availability_zone_id=availability_zone_id,
        cidr_block=cidr_block,
        default_for_az=default_for_az,
        filters=filters,
        ipv6_cidr_block=ipv6_cidr_block,
        status=status,
        vpc_id=vpc_id,
        tags=tags,
    )
    result["result"] = ret["result"]
    result["comment"] = ret["comment"]
    if result["result"] and ret["ret"]:
        result["old_state"] = ret["ret"]
        # Populate both "old_state" and "new_state" with the same data
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result



from typing import Any
from typing import Dict


async def is_target_policy_attached(
    hub, ctx, policy_id: str, target_id: str
) -> Dict[str, Any]:
    """
    Check if a organization policy is attached to a target

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        policy_id(str): The unique identifier (ID) of the policy that you want to attach to the target.
                    You can get the ID for the policy by calling the ListPolicies operation.
        target_id(str): The unique identifier (ID) of the root, OU, or account that you want to attach the policy to.


    Returns:
        {"result": True|False, "comment": "A message", "ret": None}
    """
    result = dict(comment="", result=False, ret=None)
    attached_policies_list = []
    ret_list = await hub.exec.boto3.client.organizations.list_policies_for_target(
        ctx, TargetId=target_id, Filter="SERVICE_CONTROL_POLICY"
    )
    if ret_list["result"] and ret_list["ret"].get("Policies"):
        attached_policies_list = ret_list["ret"].get("Policies")

    ret_tags_list = await hub.exec.boto3.client.organizations.list_policies_for_target(
        ctx, TargetId=target_id, Filter="TAG_POLICY"
    )
    if ret_tags_list and ret_tags_list["ret"]["Policies"]:
        attached_policies_list.extend(ret_tags_list["ret"]["Policies"])

    if attached_policies_list:
        policy_id_list = [policy.get("Id") for policy in attached_policies_list]
        if policy_id in policy_id_list:
            result["result"] = True
            result["ret"] = {"PolicyId": policy_id}
    else:
        result["comment"] = ret_list["comment"]
    return result



"""

hub.exec.boto3.client.elb.create_load_balancer
hub.exec.boto3.client.elb.delete_load_balancer
hub.exec.boto3.client.elb.describe_load_balancers
hub.exec.boto3.client.elb.delete_load_balancer_policy
hub.exec.boto3.client.elb.describe_load_balancer_attributes
hub.exec.boto3.client.elb.describe_load_balancer_policies
hub.exec.boto3.client.elb.modify_load_balancer_attributes
hub.exec.boto3.client.elb.register_instances_with_load_balancer
hub.exec.boto3.client.elb.detach_load_balancer_from_subnets
hub.exec.boto3.client.elb.enable_availability_zones_for_load_balancer
hub.exec.boto3.client.elb.disable_availability_zones_for_load_balancer
hub.exec.boto3.client.elb.deregister_instances_from_load_balancer
hub.exec.boto3.client.elb.apply_security_groups_to_load_balancer
hub.exec.boto3.client.elb.attach_load_balancer_to_subnets

"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    listeners: List[
        make_dataclass(
            "Listener",
            [
                ("Protocol", str),
                ("LoadBalancerPort", int),
                ("InstancePort", int),
                ("InstanceProtocol", str, field(default=None)),
                ("SSLCertificateId", str, field(default=None)),
            ],
        )
    ],
    availability_zones: List[str] = None,
    subnets: List[str] = None,
    security_groups: List[str] = None,
    scheme: str = None,
    instances: List[
        make_dataclass(
            "Instance",
            [
                ("InstanceId", str, field(default=None)),
            ],
        )
    ] = None,
    load_balancer_attributes: make_dataclass(
        "LoadBalancerAttributes",
        [
            (
                "CrossZoneLoadBalancing",
                make_dataclass(
                    "CrossZoneLoadBalancing",
                    [("Enabled", bool, field(default=False))],
                ),
            ),
            (
                "AccessLog",
                make_dataclass(
                    "AccessLog",
                    [
                        ("Enabled", bool, field(default=False)),
                        ("S3BucketName", str, field(default=None)),
                        ("EmitInterval", int, field(default=60)),
                        ("S3BucketPrefix", str, field(default=None)),
                    ],
                ),
            ),
            (
                "ConnectionDraining",
                make_dataclass(
                    "ConnectionDraining",
                    [
                        ("Timeout", int),
                        ("Enabled", bool, field(default=False)),
                    ],
                ),
            ),
            (
                "ConnectionSettings",
                make_dataclass(
                    "ConnectionSettings",
                    [
                        ("IdleTimeout", int),
                    ],
                ),
            ),
            (
                "AdditionalAttributes",
                List[
                    make_dataclass(
                        "AdditionalAttributes",
                        [
                            ("Key", str, field(default=None)),
                            ("Value", str, field(default=None)),
                        ],
                    )
                ],
            ),
        ],
    ) = None,
    tags: Dict[str, str] = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""

    Creates a Classic Load Balancer. You can add listeners, security groups, subnets, and tags when you create your
    load balancer, or you can add them later using CreateLoadBalancerListeners , ApplySecurityGroupsToLoadBalancer ,
    AttachLoadBalancerToSubnets and AddTags .

    To describe your current load balancers, see DescribeLoadBalancers. When you are finished with a load balancer,
    you can delete it using DeleteLoadBalancer.

    You can create up to 20 load balancers per region per account. You can request an increase for the number of load
    balancers for your account. For more information, see Limits for Your Classic Load Balancer in the
    Classic Load Balancers Guide.

    Args:
        name(Text): The name of the load balancer. This name must be unique within your set of load balancers for the
            region, must have a maximum of 32 characters, must contain only alphanumeric characters or hyphens, and
            cannot begin or end with a hyphen.
        listeners(List): The listeners.
            (Dict) --
                Information about a listener.
                Protocol (Text)
                    The load balancer transport protocol to use for routing: HTTP, HTTPS, TCP, or SSL.
                LoadBalancerPort (Integer)
                    The port on which the load balancer is listening. On EC2-VPC, you can specify any port from the
                    range 1-65535. On EC2-Classic, you can specify any port from the following list: 25, 80, 443, 465,
                    587, 1024-65535.
                InstanceProtocol (Text, Optional)
                    1. The protocol to use for routing traffic to instance: HTTP, HTTPS, TCP, or SSL. If the front-end
                    protocol is TCP or SSL, the back-end protocol must be TCP or SSL. If the front-end protocol is HTTP
                    or HTTPS, the back-end protocol must be HTTP or HTTPS.
                    2. If there is another listener with the same InstancePort whose InstanceProtocol is secure,
                    (HTTPS or SSL), the listener's InstanceProtocol must also be secure.
                    3. If there is another listener with the same InstancePort whose InstanceProtocol is HTTP or TCP,
                    the listener's InstanceProtocol must be HTTP or TCP.
                InstancePort (Integer)
                    The port on which the instance is listening.
                SSLCertificateId (Text, Optional)
                    The Amazon Resource Name (ARN) of the server certificate.
        availability_zones(List, Optional):
            One or more Availability Zones from the same region as the load balancer. You must specify at least one
            Availability Zone. You can add more Availability Zones after you create the load balancer using
            EnableAvailabilityZonesForLoadBalancer .
        subnets(List, Optional):
            The IDs of the subnets in your VPC to attach to the load balancer. Specify one subnet per Availability Zone
            specified in AvailabilityZones .
        security_groups(List, Optional):
            The IDs of the security groups to assign to the load balancer.
        scheme(Text, Optional):
            1. Given load balance type. Valid only for load balancers in a VPC. By default, Elastic Load
            Balancing creates an Internet-facing load balancer with a DNS name that resolves to public IP addresses.
            2. Specify internal to create a load balancer with a DNS name that resolves to private IP addresses.
        instances(Dict, optional):
            The IDs of the instances.
                The ID of an EC2 instance (Dict)
                    InstanceId (Text): The instance ID.
        load_balancer_attributes(Dict, Optional):
            The attributes for the load balancer.

            CrossZoneLoadBalancing(Dict): If enabled, the load balancer routes the request traffic evenly across all
                        instances regardless of the Availability Zones.
                Enabled (boolean): Specifies whether cross-zone load balancing is enabled for the load balancer.

            AccessLog(Dict, Optional): If enabled, the load balancer captures detailed information of all requests and
                        delivers the information to the Amazon S3 bucket that you specify.
                Enabled(Boolean): Specifies whether access logs are enabled for the load balancer.
                S3BucketName(Text, optional): The name of the Amazon S3 bucket where the access logs are stored.
                EmitInterval(Integer, optional): The interval for publishing the access logs. You can specify an
                    interval of either 5 minutes or 60 minutes. Default: 60 minutes
                S3BucketPrefix(Text, optional): The logical hierarchy you created for your Amazon S3 bucket, for example
                    my-bucket-prefix/prod . If the prefix is not provided, the log is placed at the root level of the
                    bucket.

            ConnectionDraining(Dict, optional): If enabled, the load balancer allows existing requests to complete
                        before the load balancer shifts traffic away from a deregistered or unhealthy instance.
                Enabled(Boolean): Specifies whether connection draining is enabled for the load balancer.
                Timeout(Integer, optional): The maximum time, in seconds, to keep the existing connections open before
                    unregistering the instances.
                ConnectionSettings(Dict): If enabled, the load balancer allows the connections to remain idle
                    (no data is sent over the connection) for the specified duration.
                IdleTimeout(Integer): The time, in seconds, that the connection is allowed to be idle (no data has been
                    sent over the connection) before it is closed by the load balancer.

            AdditionalAttributes (List, optional): Any additional attributes.
                (Dict): Information about additional load balancer attributes.
                    Key (Text): The name of the attribute. The following attribute is supported.
                        elb.http.desyncmitigationmode - Determines how the load balancer handles requests that might
                            pose a security risk to your application. The possible values are monitor , defensive ,
                            and strictest . The default is defensive .
                    Value (Text, optional): This value of the attribute.
        tags([Dict[str, str], optional): The tags to assign to the load balancer. Defaults to None.
            * Key (Text): The key of the tag.
            * Value (Text, optional): The value of the tag.

        resource_id:(Text, optional): AWS ELB load balancer name.

    Request Syntax:
    [load_balancer-name]:
      aws.elb.load_balancer.present:
      - name: 'string'
      - listeners: 'list'
      - availability_zones: 'list'
      - subnets: 'string'
      - security_groups: 'list'
      - scheme: 'string'
      - instances: 'list',
      - load_balancer_attributes: 'list',
      - tags:
          'string': 'string'
      - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

    .. code-block:: sls

        test_load-balancer_name:
          aws.elb.load_balancer.present:
            - name: my-load-balancer
            - resource_id: my-load-balancer
            - listeners:
              - InstancePort: 80
              - InstanceProtocol: HTTP
              - LoadBalancerPort: 80
              - Protocol: HTTP
            - tags:
                name: my-load-balancer
            - availability_zones
              - us-west-2a
            - subnets
              - subnet-15aaab61
            - security_groups
              - sg-a61988c378fgtyu
            - scheme: internal
            - instances:
              - i-d6f6f34tyu78ae3
              - i-207d9vbty56u717
              - i-afefvbu90ipb49b
            - load_balancer_attributes:
              - CrossZoneLoadBalancing
                Enabled: True

    """
    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)
    plan_state = {}
    before = None
    resource_updated = False
    input_map = {
        "LoadBalancerName": name,
        "Listeners": listeners,
        "Subnets": subnets,
        "AvailabilityZones": availability_zones,
        "LoadBalancerAttributes": load_balancer_attributes,
        "SecurityGroups": security_groups,
        "Scheme": scheme,
        "Tags": tags,
        "Instances": instances,
    }
    if resource_id:
        before = await hub.exec.aws.elb.load_balancer.get(
            ctx=ctx, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        update_return = await hub.tool.aws.elb.load_balancer.update(
            ctx=ctx,
            name=name,
            current_state=before["ret"],
            input_map=input_map,
            plan_state=plan_state,
            resource_id=resource_id,
        )
        result["comment"] = update_return["comment"]
        if not update_return["result"]:
            result["result"] = False
            return result
        resource_updated = bool(update_return["ret"])
        if resource_updated:
            if ctx.get("test", False):
                result["comment"] += list(
                    hub.tool.aws.comment_utils.would_update_comment(
                        resource_type="aws.elb.load_balancer", name=name
                    )
                )
            else:
                result["comment"] += list(
                    hub.tool.aws.comment_utils.update_comment(
                        resource_type="aws.elb.load_balancer", name=name
                    )
                )

        if tags is not None and tags != result["old_state"].get("tags"):
            # Update tags
            update_tags = await hub.tool.aws.elb.tag.update(
                ctx=ctx,
                resource_id=resource_id,
                old_tags=result["old_state"]["tags"],
                new_tags=tags,
            )
            result["comment"] += list(update_tags["comment"])
            result["result"] = update_tags["result"]
            resource_updated = resource_updated or bool(update_tags["ret"])
            if ctx.get("test", False) and update_tags["ret"]:
                plan_state["tags"] = update_tags["ret"]
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "listeners": listeners,
                    "availability_zones": availability_zones,
                    "subnets": subnets,
                    "security_groups": security_groups,
                    "scheme": scheme,
                    "instances": instances,
                    "load_balancer_attributes": load_balancer_attributes,
                    "tags": tags,
                    "resource_id": name,
                },
            )
            result["comment"] = list(
                hub.tool.aws.comment_utils.would_create_comment(
                    resource_type="aws.elb.load_balancer", name=name
                )
            )
            return result
        ret = await hub.tool.aws.elb.load_balancer.create(
            ctx=ctx,
            input_map=input_map,
        )
        if not ret["result"] and not ret["ret"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = list(
            hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.elb.load_balancer", name=name
            )
        )
        resource_id = name
        result["comment"] += ret["comment"]

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not result["old_state"]) or resource_updated:
        after = await hub.exec.aws.elb.load_balancer.get(
            ctx=ctx, resource_id=resource_id
        )
        if not after["result"] or not after["ret"]:
            result["result"] = False
            result["comment"] += list(after["comment"])
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    Deletes the specified load balancer.

    If you are attempting to recreate a load balancer, you must reconfigure all
    settings. The DNS name associated with a deleted load balancer are no longer usable. The name and associated DNS
    record of the deleted load balancer no longer exist and traffic sent to any of its IP addresses is no longer
    delivered to your instances.

    If the load balancer does not exist or has already been deleted, the call to DeleteLoadBalancer still succeeds.

    Args:
        name(Text): Idem name of the load balancer.
        resource_id:(Text, optional): AWS ELB load balancer name.

    Request Syntax:
        [load_balancer-name]:
          aws.elb.load_balancer.absent:
          - name: 'string'
          - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test_load-balancer_name:
                aws.elb.load_balancer.absent:
                  - name: my-load-balancer
                  - resource_id: my-load-balancer

    """
    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = list(
            hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.elb.load_balancer", name=name
            )
        )
        return result
    before = await hub.exec.aws.elb.load_balancer.get(ctx=ctx, resource_id=resource_id)
    if not before["result"]:
        result["comment"] = list(before["comment"])
        result["result"] = False
        return result

    if not before["ret"]:
        result["comment"] = list(
            hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.elb.load_balancer", name=name
            )
        )
    else:
        result["old_state"] = before["ret"]
        if ctx.get("test", False):
            result["comment"] = list(
                hub.tool.aws.comment_utils.would_delete_comment(
                    resource_type="aws.elb.load_balancer", name=name
                )
            )
            return result
        else:
            ret = await hub.exec.boto3.client.elb.delete_load_balancer(
                ctx, LoadBalancerName=resource_id
            )
            if not ret["result"]:
                result["comment"] = list(ret["comment"])
                result["result"] = False
                return result
            result["comment"] = list(
                hub.tool.aws.comment_utils.delete_comment(
                    resource_type="aws.elb.load_balancer", name=name
                )
            )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.
    This call describes all of your load balancers.

    Returns:
        Dict[str, Any]

    Examples:

    .. code-block:: bash

        $ idem describe aws.elb.load_balancer
    """
    result = {}
    ret = await hub.exec.boto3.client.elb.describe_load_balancers(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe aws.elb.load_balancer: {ret['comment']}")
        return result
    if ret["ret"].get("LoadBalancerDescriptions"):
        for load_balancer in ret["ret"]["LoadBalancerDescriptions"]:
            tags_ret = await hub.exec.boto3.client.elb.describe_tags(
                ctx, LoadBalancerNames=[load_balancer["LoadBalancerName"]]
            )
            if not tags_ret["result"]:
                # Error fetching tags. If fetching tags itself fails, just skip this and continue. Since tags are
                # optional, they need not be associated with all load balancers. But API needs to succeed.
                hub.log.warning(
                    f"Failed listing tags for aws.elb.load_balancer '{load_balancer['LoadBalancerName']}' "
                    f"with error: {tags_ret['comment']}"
                    f"Describe will skip this aws.elb.load_balancer resource and continue."
                )
                continue

            tags = []
            if tags_ret.get("ret") and tags_ret.get("ret")["TagDescriptions"]:
                tags = (tags_ret["ret"]["TagDescriptions"][0]).get("Tags")

            attributes_ret = (
                await hub.exec.boto3.client.elb.describe_load_balancer_attributes(
                    ctx, LoadBalancerName=load_balancer["LoadBalancerName"]
                )
            )
            if not attributes_ret["result"]:
                # Error fetching attributes (Attributes are optional. Not every load_balancer needs to have them.)
                hub.log.warning(
                    f"Failed listing attributes for aws.elb.load_balancer '{load_balancer.get('LoadBalancerName')}' "
                    f"with error: {attributes_ret['comment']}"
                    f"Describe will skip this aws.elb.load_balancer resource and continue."
                )
                continue

            attributes = []
            if attributes_ret.get("ret") and attributes_ret.get("ret").get(
                "LoadBalancerAttributes"
            ):
                attributes = attributes_ret["ret"]["LoadBalancerAttributes"]

            resource_converted = (
                hub.tool.aws.elb.conversion_utils.convert_raw_load_balancer_to_present(
                    raw_resource=load_balancer,
                    idem_resource_name=load_balancer["LoadBalancerName"],
                    tags=hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags),
                    attributes=attributes,
                )
            )
            result[resource_converted["resource_id"]] = {
                "aws.elb.load_balancer.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_converted.items()
                ]
            }
    return result



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


async def convert_raw_db_cluster_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    tags: Dict[str, Any] = None,
) -> Dict[str, Any]:
    """
    Convert AWS returned data structure to correct idem neptune db_cluster present state

    Args:
        hub: required for functions in hub
        raw_resource: The aws response to convert
        idem_resource_name (Text, Optional): The idem name of the resource
        tags (List, Optional): The tags of the resource. Defaults to None.

    Returns: Valid idem state for neptune db_cluster of type Dict['string', Any]
    """
    resource_id = raw_resource.get("DBClusterIdentifier")
    raw_resource["Tags"] = dict(tags)
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    # below contains everything except VpcSecurityGroups as that needs to be handled as a list
    resource_parameters = OrderedDict(
        {
            "DBClusterIdentifier": "db_cluster_identifier",
            "Engine": "engine",
            "AvailabilityZones": "availability_zones",
            "BackupRetentionPeriod": "backup_retention_period",
            "CopyTagsToSnapshot": "copy_tags_to_snapshot",
            "DatabaseName": "database_name",
            "DBClusterParameterGroup": "db_cluster_parameter_group_name",  # DBClusterParameterGroupName is provided during input and mapped to DBClusterParameterGroup in output
            "DBSubnetGroup": "db_subnet_group_name",  # DBSubnetGroupName during input is mapped to DBSubnetGroup in output
            "EngineVersion": "engine_version",
            "Port": "port",
            "PreferredBackupWindow": "preferred_backup_window",
            "PreferredMaintenanceWindow": "preferred_maintenance_window",
            "ReplicationSourceIdentifier": "replication_source_identifier",
            "Tags": "tags",
            "StorageEncrypted": "storage_encrypted",
            "KmsKeyId": "kms_key_id",
            "EnableIAMDatabaseAuthentication": "enable_iam_database_authentication",
            "EnableCloudwatchLogsExports": "enable_cloudwatch_logs_exports",
            "DeletionProtection": "deletion_protection",
            "SourceRegion": "source_region",
        }
    )
    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    # also process the VpcSecurityGroups which is a list of object to map it as list of vpc security group id strings
    if raw_resource.get("VpcSecurityGroups") is not None:
        resource_translated["vpc_security_group_ids"] = [
            vpcsecgroup["VpcSecurityGroupId"]
            for vpcsecgroup in raw_resource.get("VpcSecurityGroups")
        ]
    return resource_translated


async def get_present_db_cluster(
    hub, ctx, name: str, resource_id: str, desired_state: Dict = None
) -> Dict[str, Any]:
    """
    Get the idem present state representation of db cluster using the resource_id.

    Args:
        hub: required for functions in hub
        ctx: pop context
        name: idem name of resource,
        resource_id: db_cluster_identifier of the resource

    Returns: Valid idem state representing a neptune db cluster for the given resource id along with tags
    """
    result = dict(comment=(), result=True, present_db_cluster=None, resource_arn=None)
    if ctx.get("test", False):
        result["present_db_cluster"] = desired_state
        return result
    raw_db_cluster_result = await hub.exec.boto3.client.neptune.describe_db_clusters(
        ctx, DBClusterIdentifier=resource_id
    )
    # if we don't find db_cluster with provided resource id, fail the request
    if not (
        raw_db_cluster_result["result"]
        and raw_db_cluster_result["ret"]
        and raw_db_cluster_result["ret"].get("DBClusters")
    ):
        result["result"] = False
        result["comment"] = raw_db_cluster_result["comment"]
        return result
    resource_arn = raw_db_cluster_result["ret"].get("DBClusters")[0].get("DBClusterArn")
    result["resource_arn"] = resource_arn
    tags = await hub.tool.aws.neptune.tag.get_tags_for_resource(
        ctx, resource_arn=resource_arn
    )
    # if failure while fetching tags, fail the request
    if not tags["result"]:
        result["result"] = False
        result["comment"] = tags["comment"]
        return result
    tags = tags["ret"]
    result[
        "present_db_cluster"
    ] = await hub.tool.aws.neptune.db_cluster.convert_raw_db_cluster_to_present(
        raw_resource=raw_db_cluster_result["ret"]["DBClusters"][0],
        idem_resource_name=name,
        tags=tags,
    )
    return result


async def update_db_cluster(
    hub,
    ctx,
    name: str,
    resource_id: str,
    desired_state: Dict,
    is_created: bool,
    update_waiter_acceptors: List,
    timeout: Dict,
) -> Dict[str, Any]:
    """
    Updates the neptune db cluster represented by the provided resource_id to the desired state.
    If is_created flag is set to True, the comment in result is not updated as it is assumed that
    the resource is being updated right after it was created to apply remaining parameters.
    """
    # below is the mapping of boto3 parameters applicable for modification of neptune db_cluster in idem (tags are handled separately)
    modify_params = OrderedDict(
        {
            "ApplyImmediately": "apply_immediately",
            "BackupRetentionPeriod": "backup_retention_period",
            "DBClusterParameterGroupName": "db_cluster_parameter_group_name",
            "VpcSecurityGroupIds": "vpc_security_group_ids",
            "Port": "port",
            "MasterUserPassword": "master_user_password",
            "OptionGroupName": "option_group_name",
            "PreferredBackupWindow": "preferred_backup_window",
            "PreferredMaintenanceWindow": "preferred_maintenance_window",
            "EnableIAMDatabaseAuthentication": "enable_iam_database_authentication",
            "CloudwatchLogsExportConfiguration": "cloudwatch_logs_export_configuration",
            "EngineVersion": "engine_version",
            "AllowMajorVersionUpgrade": "allow_major_version_upgrades",
            "DBInstanceParameterGroupName": "db_instance_parameter_group_name",
            "DeletionProtection": "deletion_protection",
            "CopyTagsToSnapshot": "copy_tags_to_snapshot",
        }
    )
    result = dict(comment=(), old_state=None, new_state=None, result=True)
    current_db_cluster_result = (
        await hub.tool.aws.neptune.db_cluster.get_present_db_cluster(
            ctx=ctx, name=name, resource_id=resource_id, desired_state=desired_state
        )
    )
    if not (
        current_db_cluster_result
        and current_db_cluster_result["result"]
        and current_db_cluster_result["present_db_cluster"]
    ):
        result["result"] = False
        result["comment"] = current_db_cluster_result["comment"]
        return result

    current_state = current_db_cluster_result["present_db_cluster"]
    if not is_created:
        result["old_state"] = current_state
    resource_arn = current_db_cluster_result["resource_arn"]
    params_to_modify = {}
    plan_state = copy.deepcopy(current_state)

    # create a dict 'params_to_modify' of (non tag) raw parameter key mapped to desired value,
    # where the desired value is not none and current value does not match desired value
    hub.log.debug(f"current state before update {current_state}")
    for param_raw, param_present in modify_params.items():
        if desired_state.get(param_present) is not None and current_state.get(
            param_present
        ) != desired_state.get(param_present):
            params_to_modify[param_raw] = desired_state[param_present]
    # if we have some parameters to modify, apply it
    hub.log.debug(f"params to modify {params_to_modify}")
    if params_to_modify:
        if ctx.get("test", False):
            for key, value in params_to_modify.items():
                plan_state[modify_params.get(key)] = value
            if not is_created:
                result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.neptune.db_cluster", name=name
                )
        else:
            modify_db_cluster_result = (
                await hub.exec.boto3.client.neptune.modify_db_cluster(
                    ctx, DBClusterIdentifier=resource_id, **params_to_modify
                )
            )
            if not modify_db_cluster_result["result"]:
                result["comment"] = modify_db_cluster_result["comment"]
                result["result"] = False
                return result
            # wait for update of cluster
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=15,
                default_max_attempts=40,
                timeout_config=timeout.get("update") if timeout else None,
            )
            cluster_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
                name="ClusterModified",
                operation="DescribeDBClusters",
                argument=["DBClusters[].Status"],
                acceptors=update_waiter_acceptors,
                client=hub.tool.boto3.client.get_client(ctx, "neptune"),
            )
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "neptune",
                    "ClusterModified",
                    cluster_waiter,
                    30,
                    DBClusterIdentifier=resource_id,
                    WaiterConfig=waiter_config,
                )
            except Exception as e:
                result["comment"] = (str(e),)
                result["result"] = False
                return result
            if not is_created:
                result["comment"] = result[
                    "comment"
                ] + hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.neptune.db_cluster", name=name
                )

    # modify tags if it has changed
    old_tags = current_state["tags"]
    new_tags = desired_state["tags"]
    if new_tags is not None and new_tags != old_tags:
        # below code updates tags on AWS only if it is not a test run
        update_tags_ret = await hub.tool.aws.neptune.tag.update_tags(
            ctx=ctx,
            resource_arn=resource_arn,
            old_tags=old_tags,
            new_tags=new_tags,
        )
        if not update_tags_ret["result"]:
            result["comment"] = update_tags_ret["comment"]
            result["result"] = False
            hub.log.debug(f"Failed updating tags for aws.neptune.db_cluster '{name}'")
            return result
        # update comment if tags were updated
        result["comment"] = result["comment"] + update_tags_ret["comment"]
        if ctx.get("test", False) and update_tags_ret["ret"] is not None:
            plan_state["tags"] = update_tags_ret["ret"].get("tags")
            if not is_created:
                result["comment"] = result[
                    "comment"
                ] + hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.neptune.db_cluster", name=resource_id
                )

    # set new_state
    if ctx.get("test", False):
        result["new_state"] = plan_state
    else:
        current_db_cluster_result = (
            await hub.tool.aws.neptune.db_cluster.get_present_db_cluster(
                ctx=ctx, name=name, resource_id=resource_id
            )
        )
        if not (
            current_db_cluster_result
            and current_db_cluster_result["result"]
            and current_db_cluster_result["present_db_cluster"]
            and current_db_cluster_result["resource_arn"]
        ):
            result["result"] = False
            result["comment"] += current_db_cluster_result["comment"]
            return result
        result["new_state"] = current_db_cluster_result["present_db_cluster"]

    return result



from typing import Dict


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    load_balancer_arn: str = None,
) -> Dict:
    """
    Pass required params to get a target group resource.

    Users can specify one of the following two to fetch results: ARN of the load balancer, or an ARN of the listener.
    Here ARN of the listener takes precedence during search i.e. if both the fields from above are present, search is
    done with LB arn.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str, Optional):
            AWS ElasticLoadBalancingv2 Listener ARN to identify the listener resource.

        load_balancer_arn((str, optional):
            The Amazon Resource Name (ARN) of the ElasticLoadBalancingv2 load balancer.
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.tool.aws.elbv2.listener.search_raw(
        ctx=ctx,
        resource_id=resource_id,
        load_balancer_arn=load_balancer_arn,
    )
    if not ret["result"]:
        if "ListenerNotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.elbv2.listener", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["Listeners"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.elbv2.listener", name=name
            )
        )
        return result

    resource_id = ret["ret"]["Listeners"][0].get("ListenerArn")
    if len(ret["ret"]["Listeners"]) > 1:
        result["comment"].append(
            f"More than one aws.elbv2.listener resource was found. Use resource {resource_id}"
        )
    tags = certificates = []
    tags_ret = await hub.exec.boto3.client.elbv2.describe_tags(
        ctx, ResourceArns=[resource_id]
    )
    if not tags_ret["result"]:
        result["comment"] = list(tags_ret["comment"])
        result["result"] = False
        return result
    else:
        if tags_ret["ret"] and tags_ret.get("ret")["TagDescriptions"]:
            tags = (tags_ret["ret"]["TagDescriptions"][0]).get("Tags")

    # Listeners of type other than HTTPS & TLS don't need a certificate associated
    # with them. When this call is executed for such listeners, it returns 'False'.
    certificates_ret = await hub.exec.boto3.client.elbv2.describe_listener_certificates(
        ctx, ListenerArn=resource_id
    )
    if not certificates_ret["result"]:
        result["comment"] = list(certificates_ret["comment"])
    else:
        if certificates_ret.get("ret") and certificates_ret.get("ret").get(
            "Certificates"
        ):
            certificates = certificates_ret["ret"].get("Certificates")

    result["ret"] = hub.tool.aws.elbv2.conversion_utils.convert_raw_listener_to_present(
        raw_resource=ret["ret"]["Listeners"][0],
        idem_resource_name=name,
        tags=hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags),
        certificates=certificates,
    )
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict


async def convert_raw_acm_to_present(
    hub,
    ctx,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    result = dict(comment=(), result=True, ret=None)
    resource_id = raw_resource.get("CertificateArn")
    resource_parameters = OrderedDict(
        {
            "DomainName": "domain_name",
            "CertificateAuthorityArn": "certificate_authority_arn",
            "Status": "status",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    domain_validation_options_list = []
    # Creating domain_validation_options_list_validation list for Certificate Validation
    domain_validation_options_list_validation = []
    validation_option_list = OrderedDict(
        {
            "DomainName": "domain_name",
            "ValidationDomain": "validation_domain",
        }
    )
    if raw_resource.get("DomainValidationOptions"):
        for validation_option in raw_resource.get("DomainValidationOptions"):
            validation_option_dict = {}
            validation_option_dict_validation = {}
            for option_raw, option_present in validation_option_list.items():
                if option_raw in validation_option:
                    validation_option_dict[option_present] = validation_option.get(
                        option_raw
                    )
                    validation_option_dict_validation[
                        option_present
                    ] = validation_option.get(option_raw)

            domain_validation_options_list.append(validation_option_dict)
            if (
                not "validation_method" in resource_translated
                and "ValidationMethod" in validation_option
            ):
                resource_translated["validation_method"] = validation_option.get(
                    "ValidationMethod"
                )

            if "ResourceRecord" in validation_option:
                validation_option_dict_validation[
                    "resource_record"
                ] = validation_option.get("ResourceRecord")
            if (
                "ValidationEmails" in validation_option
                and len(validation_option.get("ValidationEmails")) > 0
            ):
                validation_option_dict_validation[
                    "validation_emails"
                ] = validation_option.get("ValidationEmails")
            if "ValidationStatus" in validation_option:
                validation_option_dict_validation[
                    "validation_status"
                ] = validation_option.get("ValidationStatus")
            domain_validation_options_list_validation.append(
                validation_option_dict_validation
            )

        resource_translated[
            "domain_validation_options"
        ] = domain_validation_options_list
        resource_translated[
            "domain_validation_options_validation"
        ] = domain_validation_options_list_validation
    if "Options" in raw_resource:
        resource_translated["options"] = {
            "CertificateTransparencyLoggingPreference": str(
                raw_resource.get("Options").get(
                    "CertificateTransparencyLoggingPreference"
                )
            )
        }

    # Cleanup SubjectAlternativeNames(As it can contain primary domain names also, so no need to keep in SubjectAlternativeNames)
    resource_translated["subject_alternative_names"] = list(
        filter(
            (raw_resource.get("DomainName")).__ne__,
            raw_resource.get("SubjectAlternativeNames"),
        )
    )
    tags = await hub.exec.boto3.client.acm.list_tags_for_certificate(
        ctx, CertificateArn=resource_id
    )
    result["result"] = tags["result"]
    if tags["result"] and tags.get("ret"):
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            tags["ret"]["Tags"]
        )
    if not tags["result"]:
        result["comment"] = result["comment"] + tags["comment"]
    result["ret"] = resource_translated
    return result


def convert_raw_acm_validation_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    result = dict(comment=(), result=True, ret=None)

    certificate_arn = raw_resource.get("CertificateArn")
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": certificate_arn,
        "certificate_arn": certificate_arn,
    }
    if raw_resource.get("DomainValidationOptions"):
        validation_record_fqdns = []
        for option in raw_resource.get("DomainValidationOptions"):
            if option.get("ResourceRecord"):
                if option.get("ResourceRecord").get("Name"):
                    validation_record_fqdns.append(
                        option.get("ResourceRecord").get("Name")
                    )
        resource_translated["validation_record_fqdns"] = validation_record_fqdns

    result["ret"] = resource_translated
    return result



import inspect
import json
from typing import Any
from typing import Dict
from typing import List


"""
ATTENTION: Functions in this util file are highly controversial and we don't encourage use any of these functions.
1. These functions give developers a misconception that they are suitable for comparing certain data structure.
In reality, they are not. They all have limitations and only work on certain scope of data structures. Data structure
such as List and Dict can contain nested data structures and these functions will certainly not do the job well on
nested data structure.
2. There are mature modules if we have to compare two complex data structure, for example, dict_tool, deepdiff.
The functions here are re-inventing the wheel.

Hence, think about twice before you decide to use any of the following functions. There should be no new util function
added to this file and the existing functions should be deprecated at some point.
"""


def are_lists_identical(hub, list1: List, list2: List) -> bool:
    """
    Compare two lists and logs the difference.
    :param list1: first list.
    :param list2: second list.
    :return: true if there is no difference between both lists.
    :raises exception if one of the list  is not of type list
    """
    if (list1 is None or len(list1) == 0) and (list2 is None or len(list2) == 0):
        return True
    if list1 is None or len(list1) == 0 or list2 is None or len(list2) == 0:
        return False

    for l in [list1, list2]:
        if not isinstance(l, List):
            raise TypeError(
                f"Expecting lists to compare. This is expected to be of type List: '{l}'"
            )

    diff = [i for i in list1 + list2 if i not in list1 or i not in list2]
    result = len(diff) == 0
    if not result:
        hub.log.debug(f"There are {len(diff)} differences:\n{diff[:5]}")
    return result


def standardise_json(hub, value: str or Dict, sort_keys: bool = True) -> str:
    # Format json string or dictionary
    if value is None or value is inspect._empty or not value:
        return None

    if isinstance(value, str) and len(value) > 0:
        json_dict = json.loads(value)
    elif isinstance(value, Dict):
        json_dict = value
    else:
        raise TypeError(
            f"Expecting string or dictionary. This value has the wrong type: '{value}'"
        )

    return json.dumps(
        _sorting(json_dict) if sort_keys else json_dict,
        separators=(", ", ": "),
        sort_keys=sort_keys,
    )


def is_json_identical(hub, struct1: str, struct2: str):
    return _sorting(json.loads(struct1)) == _sorting(json.loads(struct2))


def _sorting(obj):
    if isinstance(obj, dict):
        return {k: _sorting(v) for k, v in sorted(obj.items())}
    if isinstance(obj, list):
        if all(isinstance(v, str) for v in obj):
            return sorted(obj)
        else:
            return [_sorting(v) for i, v in sorted(enumerate(obj))]
    return obj


def compare_dicts(hub, source_dict: Dict[str, Any], target_dict: Dict[str, Any]):
    """
    This functions helps in comparing two dicts.
    It compares each key value in both the dicts and return true or false based on the comparison

    Returns:
        {True|False}

    """

    for key, value in source_dict.items():
        if key in target_dict:
            if isinstance(source_dict[key], dict):
                if not compare_dicts(hub, source_dict[key], target_dict[key]):
                    return False
            elif value != target_dict[key]:
                return False
        else:
            return False
    return True
